from tarfile import TarInfo

import pytest

from dvcx.lib.file import File
from dvcx.lib.webdataset import (
    CoreFileDuplicationError,
    CoreFileNotFoundError,
    UnknownFileExtensionError,
    WebDataset,
)
from dvcx.lib.webdataset_laion import WDSLaion


class MockTarInfo(TarInfo):
    def __init__(self, name, content=b"", size=0, offset=0):
        super().__init__(name)
        self._content = content.encode() if isinstance(content, str) else content
        self.size = size
        self.offset = offset

    def isfile(self):
        return True


class MockTar:
    def __init__(self, members: list[MockTarInfo]):
        self._members = members

    def getmembers(self):
        return self._members

    def extractfile(self, tar_info: MockTarInfo):
        class TmpReader:
            def read(self):
                return tar_info._content

        return TmpReader()


def test_webdataset_basic():
    tar_file = File(name="nnn.tar")
    tar = MockTar(
        [
            MockTarInfo("01.jpg"),
            MockTarInfo("01.json", b'{"uid": "5678"}'),
            MockTarInfo("64.jpg"),
            MockTarInfo("64.json", b"{}"),
        ]
    )

    groups = list(WebDataset().get_tar_groups(tar_file, tar, ["jpg"], WDSLaion))

    assert len(groups) == 2

    (file01, laion01), (file64, laion64) = groups

    assert file01.name == "01.jpg"
    assert file01.parent == tar_file.name
    assert file01.location is not None
    assert isinstance(file01.location, list)
    assert len(file01.location) > 0

    parent = file01.location[0].get("parent", None)
    assert parent is not None
    parent_file = File(**parent)
    assert parent_file == tar_file

    assert file64.name == "64.jpg"
    assert file64.parent == tar_file.name

    assert laion01.txt is None
    assert laion01.json.uid == "5678"

    assert laion64.json.uid == ""


def test_webdataset_empty():
    stream = File(name="nnn.tar")
    tar = MockTar([])

    groups = list(WebDataset().get_tar_groups(stream, tar, ["jpg"], WDSLaion))

    assert len(groups) == 0


def test_webdataset_missing_core_files():
    stream = File(name="nnn.tar")
    tar = MockTar(
        [
            MockTarInfo("01.txt"),
            MockTarInfo("01.json", b'{"uid": "5678"}'),
            MockTarInfo("64.txt"),
            MockTarInfo("64.json", b"{}"),
        ]
    )

    with pytest.raises(CoreFileNotFoundError):
        list(WebDataset().get_tar_groups(stream, tar, ["NONSENSE"], WDSLaion))


def test_webdataset_single_file_per_group():
    stream = File(name="nnn.tar")
    tar = MockTar(
        [
            MockTarInfo("01.jpg"),
            MockTarInfo("64.jpg"),
            MockTarInfo("64.json", b"{}"),
            MockTarInfo("03.jpg"),
        ]
    )

    groups = list(WebDataset().get_tar_groups(stream, tar, ["jpg"], WDSLaion))
    assert len(groups) == 3


def test_webdataset_multiple_core_extensions():
    stream = File(name="nnn.tar")
    tar = MockTar(
        [
            MockTarInfo("01.png"),
            MockTarInfo("64.png"),
            MockTarInfo("64.json", b"{}"),
            MockTarInfo("03.jpg"),
        ]
    )

    groups = list(WebDataset().get_tar_groups(stream, tar, ["jpg", "png"], WDSLaion))
    assert len(groups) == 3


def test_webdataset_core_file_duplication():
    stream = File(name="nnn.tar")
    tar = MockTar(
        [
            MockTarInfo("01.png"),
            MockTarInfo("01.jpg"),
            MockTarInfo("64.json", b"{}"),
            MockTarInfo("03.jpg"),
        ]
    )

    with pytest.raises(CoreFileDuplicationError):
        list(WebDataset().get_tar_groups(stream, tar, ["jpg", "png"], WDSLaion))


def test_webdataset_unknown_file_type():
    stream = File(name="nnn.tar")
    tar = MockTar([MockTarInfo("01.QQQQ")])

    with pytest.raises(UnknownFileExtensionError):
        list(WebDataset().get_tar_groups(stream, tar, ["jpg", "png"], WDSLaion))



# pip install Pillow

import os

from dvcx.lib.gpt4_vision import DescribeImage
from dvcx.query import C, DatasetQuery

OPENAI_API_KEY = os.getenv("OPENAI_API_KEY", "")

source = "gcs://dvcx-datalakes/dogs-and-cats/"

if __name__ == "__main__":
    results = (
        DatasetQuery(
            source,
            anon=True,
        )
        .filter(C.name.glob("cat*.jpg"))
        .limit(10)
        .add_signals(
            DescribeImage(
                key=OPENAI_API_KEY, max_tokens=300, prompt="What is in this image?"
            ),
            parallel=-1,
        )
        .select("source", "parent", "name", "description", "error")
        .results()
    )
    print(*results, sep="\n")



import shutil
import tempfile

from unstructured.partition.auto import partition
from unstructured.staging.base import convert_to_dataframe

from dvcx.lib.udf import Mapper
from dvcx.query import Stream
from dvcx.sql.types import JSON, String


class PartitionObject(Mapper):
    def __init__(self):
        super().__init__(
            [
                Stream(),
            ],
            {
                "elements": JSON,
                "title": String,
                "text": String,
                "error": String,
            },
        )

    def encode_object(self, raw):
        fname = str(raw).replace(">", "").replace("<", "")
        output = tempfile.TemporaryFile()
        shutil.copyfileobj(raw, output)
        elements = partition(file=output, metadata_filename=fname)
        output.close()
        return elements

    def __call__(self, stream):
        with stream:
            elements = self.encode_object(stream)

        title = str(elements[0])
        text = "\n\n".join([str(el) for el in elements])
        df = convert_to_dataframe(elements)
        return (df.to_json(), title, text, "")



import logging
from collections.abc import Iterator, Sequence
from typing import TYPE_CHECKING, Any, Optional

from torch import float32
from torch.distributed import get_rank, get_world_size
from torch.utils.data import IterableDataset, get_worker_info

from dvcx.catalog import Catalog, get_catalog
from dvcx.lib.dataset import Dataset

if TYPE_CHECKING:
    from torchvision.transforms.v2 import Transform

    from dvcx.lib.feature_types import FeatureLike


logger = logging.getLogger("dvcx")


try:
    from torchvision.transforms import v2

    DEFAULT_TRANSFORM = v2.Compose([v2.ToImage(), v2.ToDtype(float32, scale=True)])
except ImportError:
    logger.warning("Missing dependency torchvision for computer vision transforms.")
    DEFAULT_TRANSFORM = None


class PytorchDataset(IterableDataset):
    def __init__(
        self,
        fr_classes: Sequence["FeatureLike"],
        name: str,
        version: Optional[int] = None,
        catalog: Optional["Catalog"] = None,
        transform: Optional["Transform"] = DEFAULT_TRANSFORM,
        num_samples: int = 0,
    ):
        """
        Pytorch IterableDataset that streams DVCx datasets.

        Args:
            fr_classes: Feature-like classes from DVCx dataset to stream.
            name (str): Name of DVCx dataset to stream.
            version (int): Version of DVCx dataset to stream.
            catalog (Catalog): DVCx catalog to which dataset belongs.
            transform (Transform): Torchvision v2 transforms to apply to the dataset.
            num_samples (int): Number of random samples to draw for each epoch.
                This argument is ignored if `num_samples=0` (the default).
        """
        self.fr_classes = fr_classes
        self.name = name
        self.version = version
        self.transform = transform
        self.num_samples = num_samples
        if catalog is None:
            catalog = get_catalog()
        self._init_catalog(catalog)

    def _init_catalog(self, catalog: "Catalog"):
        # For compatibility with multiprocessing,
        # we can only store params in __init__(), as Catalog isn't picklable
        # see https://github.com/iterative/dvcx/issues/954
        self._idgen_params = catalog.id_generator.clone_params()
        self._ms_params = catalog.metastore.clone_params()
        self._wh_params = catalog.warehouse.clone_params()
        self._catalog_params = catalog.get_init_params()
        self.catalog: Optional[Catalog] = None

    def _get_catalog(self) -> "Catalog":
        idgen_cls, idgen_args, idgen_kwargs = self._idgen_params
        idgen = idgen_cls(*idgen_args, **idgen_kwargs)
        ms_cls, ms_args, ms_kwargs = self._ms_params
        ms = ms_cls(*ms_args, **ms_kwargs)
        wh_cls, wh_args, wh_kwargs = self._wh_params
        wh = wh_cls(*wh_args, **wh_kwargs)
        return Catalog(idgen, ms, wh, **self._catalog_params)

    def __iter__(self) -> Iterator[Any]:
        if self.catalog is None:
            self.catalog = self._get_catalog()
        total_rank, total_workers = self.get_rank_and_workers()
        ds = Dataset(name=self.name, version=self.version, catalog=self.catalog)
        if self.num_samples > 0:
            ds = ds.sample(self.num_samples)
        ds = ds.chunk(total_rank, total_workers)
        stream = ds.get_values(*self.fr_classes)
        for row in stream:
            # Apply transforms
            if self.transform:
                try:
                    row = self.transform(row)
                except ValueError:
                    logger.warning("Skipping transform due to unsupported data types.")
                    self.transform = None
            yield row

    @staticmethod
    def get_rank_and_workers() -> tuple[int, int]:
        """Get combined rank and number of workers across all nodes."""
        try:
            world_rank = get_rank()
            world_size = get_world_size()
        except (RuntimeError, ValueError):
            world_rank = 0
            world_size = 1
        worker_info = get_worker_info()
        if worker_info:
            worker_rank = worker_info.id
            num_workers = worker_info.num_workers
        else:
            worker_rank = 0
            num_workers = 1
        total_workers = world_size * num_workers
        total_rank = world_rank * num_workers + worker_rank
        return total_rank, total_workers



try:
    import numpy as np
    import torch
    from PIL import Image, ImageOps, UnidentifiedImageError
    from transformers import (
        AutoProcessor,
        Blip2ForConditionalGeneration,
        Blip2Processor,
        LlavaForConditionalGeneration,
    )
except ImportError as exc:
    raise ImportError(
        "Missing dependencies for computer vision:\n"
        "To install run:\n\n"
        "  pip install 'dvcx[cv]'\n"
    ) from exc


from dvcx.query import Object, udf
from dvcx.sql.types import String

DEFAULT_FIT_BOX = (500, 500)


def encode_image(raw):
    try:
        img = Image.open(raw)
    except UnidentifiedImageError:
        return None
    img.load()
    img = img.convert("RGB")
    return ImageOps.fit(img, DEFAULT_FIT_BOX)


def infer_dtype(device):
    if device == "cpu":
        return torch.float32
    return torch.float16


@udf(
    params=(Object(encode_image),),  # Columns consumed by the UDF.
    output={
        "description": String,
        "error": String,
    },  # Signals being returned by the UDF.
    batch=64,
    method="describe",
)
class BLIP2describe:
    def __init__(self, device="cpu", model="Salesforce/blip2-opt-2.7b", max_tokens=300):
        self.torch_dtype = infer_dtype(device)
        self.processor = Blip2Processor.from_pretrained(model)
        self.model = Blip2ForConditionalGeneration.from_pretrained(
            model, torch_dtype=self.torch_dtype
        )
        self.device = device
        self.model.to(device)
        self.max_tokens = max_tokens

    def describe(self, imgs):
        images = np.squeeze(np.asarray(imgs))
        inputs = self.processor(images=images, return_tensors="pt").to(
            self.device, self.torch_dtype
        )

        generated_ids = self.model.generate(**inputs, max_new_tokens=self.max_tokens)
        generated_text = self.processor.batch_decode(
            generated_ids, skip_special_tokens=True
        )
        return [(desc.strip(), "") for desc in generated_text]


@udf(
    params=(Object(encode_image),),  # Columns consumed by the UDF.
    output={
        "description": String,
        "error": String,
    },  # Signals being returned by the UDF.
    batch=16,
    method="describe",
)
class LLaVAdescribe:
    def __init__(self, device="cpu", model="llava-hf/llava-1.5-7b-hf", max_tokens=300):
        self.device = device
        self.torch_dtype = infer_dtype(device)
        self.processor = AutoProcessor.from_pretrained(model)
        self.model = LlavaForConditionalGeneration.from_pretrained(
            model, torch_dtype=self.torch_dtype, low_cpu_mem_usage=True
        )
        self.model.to(device)
        self.max_tokens = max_tokens
        self.prompt = "USER: <image>\nDescribe this picture\nASSISTANT:"

    def describe(self, imgs):
        images = np.squeeze(np.asarray(imgs))
        inputs = self.processor(
            text=[self.prompt] * len(imgs), images=images, return_tensors="pt"
        ).to(self.device, self.torch_dtype)

        generated_ids = self.model.generate(**inputs, max_new_tokens=self.max_tokens)
        generated_text = self.processor.batch_decode(
            generated_ids, skip_special_tokens=True
        )
        return [(desc.split("ASSISTANT:")[-1].strip(), "") for desc in generated_text]



import os
import os.path
import uuid
from collections.abc import Generator
from pathlib import PosixPath

import attrs
import pytest
import pytest_servers.exceptions
import sqlalchemy
from pytest import MonkeyPatch, TempPathFactory
from upath.implementations.cloud import CloudPath

from dvcx.catalog import Catalog
from dvcx.catalog.loader import get_id_generator, get_metastore, get_warehouse
from dvcx.client.local import FileClient
from dvcx.data_storage.sqlite import (
    SQLiteDatabaseEngine,
    SQLiteIDGenerator,
    SQLiteMetastore,
    SQLiteWarehouse,
)
from dvcx.dataset import DatasetRecord, DatasetRow
from dvcx.query.session import Session
from dvcx.utils import DVCXDir, get_env_list

from .utils import DEFAULT_TREE, get_simple_ds_query, instantiate_tree

DEFAULT_DVCX_BIN = "dvcx"
DEFAULT_DVCX_GIT_REPO = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

collect_ignore = ["setup.py"]


@pytest.fixture(scope="session")
def monkeypatch_session() -> Generator[MonkeyPatch, None, None]:
    """
    Like monkeypatch, but for session scope.
    """
    mpatch = pytest.MonkeyPatch()
    yield mpatch
    mpatch.undo()


@pytest.fixture(scope="session", autouse=True)
def clean_environment(
    monkeypatch_session: MonkeyPatch,
    tmp_path_factory: TempPathFactory,
) -> None:
    """
    Make sure we have a clean environment and won't write to userspace.
    """
    working_dir = str(tmp_path_factory.mktemp("default_working_dir"))
    monkeypatch_session.chdir(working_dir)
    monkeypatch_session.delenv(DVCXDir.ENV_VAR, raising=False)

    def default_root(cls: type[DVCXDir]) -> str:
        return os.path.join(working_dir, cls.DEFAULT)

    monkeypatch_session.setattr(
        "dvcx.utils.DVCXDir.default_root",
        classmethod(default_root),
    )


@pytest.fixture
def sqlite_db():
    return SQLiteDatabaseEngine.from_db_file(":memory:")


def cleanup_sqlite_db(
    db: SQLiteDatabaseEngine,
    cleanup_tables: list[str],
):
    # Wipe the DB after the test
    # Using new connection to check that the DB isn't locked
    tables = db.execute_str(
        "SELECT name FROM sqlite_master WHERE type='table'"
    ).fetchall()

    # removing in reversed order because of foreign keys
    for table in reversed(cleanup_tables):
        db.execute_str(f"DROP TABLE IF EXISTS '{table}'")

    for (table,) in tables:
        name = table.replace("'", "''")
        db.execute_str(f"DROP TABLE IF EXISTS '{name}'")

    # Close the connection so that the SQLite file is no longer open, to avoid
    # pytest throwing: OSError: [Errno 24] Too many open files
    db.close()


@pytest.fixture
def id_generator():
    if os.environ.get("DVCX_ID_GENERATOR"):
        _id_generator = get_id_generator()
    else:
        db = SQLiteDatabaseEngine.from_db_file(":memory:")
        _id_generator = SQLiteIDGenerator(db)
    yield _id_generator
    _id_generator.cleanup_for_tests()

    if not os.environ.get("DVCX_ID_GENERATOR"):
        # Close the connection so that the SQLite file is no longer open, to avoid
        # pytest throwing: OSError: [Errno 24] Too many open files
        _id_generator.db.close()


@pytest.fixture
def metastore(id_generator):
    if os.environ.get("DVCX_METASTORE"):
        _metastore = get_metastore(id_generator)
        yield _metastore

        _metastore.cleanup_for_tests()
    else:
        _metastore = SQLiteMetastore(id_generator, db_file=":memory:")
        yield _metastore

        cleanup_sqlite_db(_metastore.db.clone(), _metastore.default_table_names)
        Session.cleanup_for_tests()

        # Close the connection so that the SQLite file is no longer open, to avoid
        # pytest throwing: OSError: [Errno 24] Too many open files
        _metastore.db.close()


def check_temp_tables_cleaned_up(original_warehouse):
    """Ensure that temporary tables are cleaned up."""
    warehouse = original_warehouse.clone()
    assert [
        t
        for t in sqlalchemy.inspect(warehouse.db.engine).get_table_names()
        if t.startswith(
            (warehouse.UDF_TABLE_NAME_PREFIX, warehouse.TMP_TABLE_NAME_PREFIX)
        )
    ] == []


@pytest.fixture
def warehouse(id_generator, metastore):
    if os.environ.get("DVCX_WAREHOUSE"):
        _warehouse = get_warehouse(id_generator)
        yield _warehouse
        check_temp_tables_cleaned_up(_warehouse)

        _warehouse.cleanup_for_tests()
    else:
        _warehouse = SQLiteWarehouse(id_generator, db_file=":memory:")
        yield _warehouse
        check_temp_tables_cleaned_up(_warehouse)

        cleanup_sqlite_db(_warehouse.db.clone(), metastore.default_table_names)

        # Close the connection so that the SQLite file is no longer open, to avoid
        # pytest throwing: OSError: [Errno 24] Too many open files
        _warehouse.db.close()


@pytest.fixture
def catalog(id_generator, metastore, warehouse):
    return Catalog(id_generator=id_generator, metastore=metastore, warehouse=warehouse)


@pytest.fixture
def id_generator_tmpfile(tmp_path):
    if os.environ.get("DVCX_ID_GENERATOR"):
        _id_generator = get_id_generator()
    else:
        db = SQLiteDatabaseEngine.from_db_file(tmp_path / "test.db")
        _id_generator = SQLiteIDGenerator(db)
    yield _id_generator
    _id_generator.cleanup_for_tests()

    if not os.environ.get("DVCX_ID_GENERATOR"):
        # Close the connection so that the SQLite file is no longer open, to avoid
        # pytest throwing: OSError: [Errno 24] Too many open files
        _id_generator.db.close()


@pytest.fixture
def metastore_tmpfile(tmp_path, id_generator_tmpfile):
    if os.environ.get("DVCX_METASTORE"):
        _metastore = get_metastore(id_generator_tmpfile)
        yield _metastore

        _metastore.cleanup_for_tests()
    else:
        _metastore = SQLiteMetastore(id_generator_tmpfile, db_file=tmp_path / "test.db")
        yield _metastore

        cleanup_sqlite_db(_metastore.db.clone(), _metastore.default_table_names)
        Session.cleanup_for_tests()

        # Close the connection so that the SQLite file is no longer open, to avoid
        # pytest throwing: OSError: [Errno 24] Too many open files
        _metastore.db.close()


@pytest.fixture
def warehouse_tmpfile(tmp_path, id_generator_tmpfile, metastore_tmpfile):
    if os.environ.get("DVCX_WAREHOUSE"):
        _warehouse = get_warehouse(id_generator_tmpfile)
        yield _warehouse
        check_temp_tables_cleaned_up(_warehouse)

        _warehouse.cleanup_for_tests()
    else:
        _warehouse = SQLiteWarehouse(id_generator_tmpfile, db_file=tmp_path / "test.db")
        yield _warehouse
        check_temp_tables_cleaned_up(_warehouse)

        cleanup_sqlite_db(_warehouse.db.clone(), metastore_tmpfile.default_table_names)

        # Close the connection so that the SQLite file is no longer open, to avoid
        # pytest throwing: OSError: [Errno 24] Too many open files
        _warehouse.db.close()


@pytest.fixture
def tmp_dir(tmp_path_factory, monkeypatch):
    dpath = tmp_path_factory.mktemp("dvcx-test")
    monkeypatch.chdir(dpath)
    return dpath


def pytest_addoption(parser):
    parser.addoption(
        "--dvcx-bin",
        type=str,
        default=DEFAULT_DVCX_BIN,
        help="Path to dvcx binary",
    )

    parser.addoption(
        "--dvcx-revs",
        type=str,
        help="Comma-separated list of DVCx revisions to test (overrides `--dvcx-bin`)",
    )

    parser.addoption(
        "--dvcx-git-repo",
        type=str,
        default=DEFAULT_DVCX_GIT_REPO,
        help="Path or url to dvcx git repo",
    )
    parser.addoption(
        "--azure-connection-string",
        type=str,
        help=(
            "Connection string to run tests against a real, versioned "
            "Azure storage account"
        ),
    )


class DVCXTestConfig:
    def __init__(self):
        self.dvcx_bin = DEFAULT_DVCX_BIN
        self.dvcx_revs = None
        self.dvcx_git_repo = DEFAULT_DVCX_GIT_REPO


@pytest.fixture(scope="session")
def test_config(request):
    return request.config.dvcx_config


def pytest_configure(config):
    config.dvcx_config = DVCXTestConfig()

    config.dvcx_config.dvcx_bin = config.getoption("--dvcx-bin")
    config.dvcx_config.dvcx_revs = config.getoption("--dvcx-revs")
    config.dvcx_config.dvcx_git_repo = config.getoption("--dvcx-git-repo")


@pytest.fixture(scope="session", params=[DEFAULT_TREE])
def tree(request):
    return request.param


@attrs.define
class CloudServer:
    kind: str
    src: CloudPath
    client_config: dict[str, str]

    @property
    def src_uri(self):
        if self.kind == "file":
            return self.src.as_uri()
        return str(self.src).rstrip("/")


def make_cloud_server(src_path, cloud_type, tree):
    fs = src_path.fs
    if cloud_type == "s3":
        endpoint_url = fs.client_kwargs["endpoint_url"]
        client_config = {"aws_endpoint_url": endpoint_url}
    elif cloud_type == "gcs":
        endpoint_url = fs._endpoint
        client_config = {"endpoint_url": endpoint_url}
    elif cloud_type == "azure":
        client_config = fs.storage_options.copy()
    elif cloud_type == "file":
        client_config = {}
    else:
        raise ValueError(f"invalid cloud_type: {cloud_type}")

    instantiate_tree(src_path, tree)
    return CloudServer(kind=cloud_type, src=src_path, client_config=client_config)


@attrs.define
class CloudTestCatalog:
    server: CloudServer
    working_dir: PosixPath
    catalog: Catalog

    @property
    def src(self):
        return self.server.src

    @property
    def src_uri(self):
        return self.server.src_uri

    @property
    def storage_uri(self):
        if self.server.kind == "file":
            return FileClient.root_path().as_uri()
        return self.server.src_uri

    @property
    def partial_path(self):
        if self.server.kind == "file":
            _, rel_path = FileClient.split_url(self.src_uri)
            return rel_path
        return ""

    @property
    def client_config(self):
        return self.server.client_config


@pytest.fixture(scope="session", params=["file", "s3", "gcs", "azure"])
def cloud_type(request):
    return request.param


@pytest.fixture(scope="session", params=[False, True])
def version_aware(request):
    return request.param


@pytest.fixture(scope="session")
def cloud_server(request, tmp_upath_factory, cloud_type, version_aware, tree):
    # DVCX_TEST_SKIP_MISSING_REMOTES can be set to a comma-separated list
    # of remotes to skip tests for if unavailable or "all" to skip all
    # unavailable remotes:
    #  DVCX_TEST_SKIP_MISSING_REMOTES=azure,gcs
    #  DVCX_TEST_SKIP_MISSING_REMOTES=all
    skip_missing_remotes = set(get_env_list("DVCX_TEST_SKIP_MISSING_REMOTES", []))
    try:
        if cloud_type == "azure" and version_aware:
            if conn_str := request.config.getoption("--azure-connection-string"):
                src_path = tmp_upath_factory.azure(conn_str)
            else:
                pytest.skip("Can't test versioning with Azure")
        elif cloud_type == "file":
            if version_aware:
                pytest.skip("Local storage can't be versioned")
            else:
                src_path = tmp_upath_factory.mktemp("local")
        else:
            src_path = tmp_upath_factory.mktemp(cloud_type, version_aware=version_aware)
    except pytest_servers.exceptions.RemoteUnavailable as exc:
        if "all" in skip_missing_remotes or cloud_type in skip_missing_remotes:
            pytest.skip(str(exc))
        raise

    return make_cloud_server(src_path, cloud_type, tree)


@pytest.fixture()
def dvcx_job_id(monkeypatch):
    job_id = uuid.uuid4().hex
    monkeypatch.setenv("DVCX_JOB_ID", job_id)


@pytest.fixture
def cloud_server_credentials(cloud_server, monkeypatch):
    if cloud_server.kind == "s3":
        cfg = cloud_server.src.fs.client_kwargs
        try:
            monkeypatch.delenv("AWS_PROFILE")
        except KeyError:
            pass
        monkeypatch.setenv("AWS_ACCESS_KEY_ID", cfg.get("aws_access_key_id"))
        monkeypatch.setenv("AWS_SECRET_ACCESS_KEY", cfg.get("aws_secret_access_key"))
        monkeypatch.setenv("AWS_SESSION_TOKEN", cfg.get("aws_session_token"))
        monkeypatch.setenv("AWS_DEFAULT_REGION", cfg.get("region_name"))


def get_cloud_test_catalog(cloud_server, tmp_path, id_generator, metastore, warehouse):
    cache_dir = tmp_path / ".dvcx" / "cache"
    cache_dir.mkdir(parents=True)
    tmpfile_dir = tmp_path / ".dvcx" / "tmp"
    tmpfile_dir.mkdir()

    catalog = Catalog(
        id_generator=id_generator,
        metastore=metastore,
        warehouse=warehouse,
        cache_dir=str(cache_dir),
        tmp_dir=str(tmpfile_dir),
        client_config=cloud_server.client_config,
    )
    return CloudTestCatalog(server=cloud_server, working_dir=tmp_path, catalog=catalog)


@pytest.fixture
def cloud_test_catalog(
    cloud_server,
    cloud_server_credentials,
    tmp_path,
    id_generator,
    metastore,
    warehouse,
):
    return get_cloud_test_catalog(
        cloud_server, tmp_path, id_generator, metastore, warehouse
    )


@pytest.fixture
def cloud_test_catalog_tmpfile(
    cloud_server,
    cloud_server_credentials,
    tmp_path,
    id_generator_tmpfile,
    metastore_tmpfile,
    warehouse_tmpfile,
):
    return get_cloud_test_catalog(
        cloud_server,
        tmp_path,
        id_generator_tmpfile,
        metastore_tmpfile,
        warehouse_tmpfile,
    )


@pytest.fixture
def listed_bucket(cloud_test_catalog):
    list(cloud_test_catalog.catalog.ls([cloud_test_catalog.src_uri], fields=["name"]))


@pytest.fixture
def dogs_dataset(listed_bucket, cloud_test_catalog):
    name = uuid.uuid4().hex
    catalog = cloud_test_catalog.catalog
    src_uri = cloud_test_catalog.src_uri
    dataset = catalog.create_dataset_from_sources(
        name, [f"{src_uri}/dogs/*"], recursive=True
    )
    return catalog.update_dataset(
        dataset, {"description": "dogs dataset", "labels": ["dogs", "dataset"]}
    )


@pytest.fixture
def cats_dataset(listed_bucket, cloud_test_catalog):
    name = uuid.uuid4().hex
    catalog = cloud_test_catalog.catalog
    src_uri = cloud_test_catalog.src_uri
    dataset = catalog.create_dataset_from_sources(
        name, [f"{src_uri}/cats/*"], recursive=True
    )
    return catalog.update_dataset(
        dataset, {"description": "cats dataset", "labels": ["cats", "dataset"]}
    )


@pytest.fixture
def simple_ds_query(cloud_test_catalog):
    return get_simple_ds_query(
        path=cloud_test_catalog.src_uri, catalog=cloud_test_catalog.catalog
    )


@pytest.fixture
def dataset_record():
    return DatasetRecord(
        id=1,
        name=f"ds_{uuid.uuid4().hex}",
        description="",
        labels=[],
        versions=[],
        shadow=False,
        status=1,
        custom_column_types={},
    )


@pytest.fixture
def dataset_rows():
    int_example = 25
    return [
        DatasetRow.from_dict(
            {
                "id": i,
                "location": "",
                "source": "s3://my-bucket",
                "dir_type": 0,
                "vtype": "",
                "parent": "input/text_emd_1m",
                "version": "7e589b7d-382c-49a5-931f-2b999c930c5e",
                "is_latest": True,
                "name": f"dql_1m_meta_text_emd.parquet_3_{i}_0.snappy.parquet",
                "etag": f"72b35c8e9b8eed1636c91eb94241c2f8-{i}",
                "owner_id": "owner",
                "owner_name": "aws-iterative-sandbox",
                "last_modified": "2024-02-23T10:42:31.842944+00:00",
                "size": 49807360,
                "random": 12123123123,
                "custom": {
                    "int_col": 5,
                    "int_col_32": 5,
                    "int_col_64": 5,
                    "float_col": 0.5,
                    "float_col_32": 0.5,
                    "float_col_64": 0.5,
                    "array_col": [0.5],
                    "array_col_nested": [[0.5]],
                    "array_col_32": [0.5],
                    "array_col_64": [0.5],
                    "string_col": "a string",
                    "bool_col": True,
                    "json_col": '{"a": 1}',
                    "binary_col": int_example.to_bytes(2, "big"),
                },
            }
        )
        for i in range(19)
    ]



import uuid

from tabulate import tabulate

from dvcx.query import C, DatasetQuery, udf
from dvcx.sql.types import Int

# To install script dependencies: pip install tabulate


# Define the UDF:
@udf(
    ("name",),  # Columns consumed by the UDF.
    {
        "name_len": Int
    },  # Signals being returned by the UDF, with the signal name and type.
)
def name_len(name):
    if name.endswith(".json"):
        return (-1,)
    return (len(name),)


if __name__ == "__main__":
    ds_name = uuid.uuid4().hex
    print(f"Saving to dataset: {ds_name}")
    # Save as a new dataset
    DatasetQuery(
        path="gcs://dvcx-datalakes/dogs-and-cats/",
        anon=True,
    ).filter(C.name.glob("*cat*")).add_signals(name_len).save(ds_name)

    # Output the contents of the new dataset.
    print(
        tabulate(
            DatasetQuery(name=ds_name)
            .order_by(C.parent, C.name)
            .limit(10)
            .select(C.source, C.parent, C.name, C.size, C.name_len)
            .results()
        )
    )



from sqlalchemy.sql.functions import GenericFunction

from dvcx.sql.types import Array, Int64, String
from dvcx.sql.utils import compiler_not_implemented


class length(GenericFunction):  # noqa: N801
    type = Int64()
    package = "string"
    name = "length"
    inherit_cache = True


class split(GenericFunction):  # noqa: N801
    type = Array(String())
    package = "string"
    name = "split"
    inherit_cache = True


compiler_not_implemented(length)
compiler_not_implemented(split)



import sys
from time import sleep

from dvcx.query import C, DatasetQuery, udf
from dvcx.sql.types import Int

if sys.platform == "win32":
    # This is needed for this process to accept a Ctrl-C event in Windows,
    # when run under pytest as a subprocess.
    # This is not needed when running normally from the command line.
    import ctypes

    kernel32 = ctypes.WinDLL("kernel32", use_last_error=True)

    if not kernel32.SetConsoleCtrlHandler(None, False):
        print("SetConsoleCtrlHandler error: ", ctypes.get_last_error(), file=sys.stderr)


# Define the UDF:
@udf(
    ("name",),  # Columns consumed by the UDF.
    {
        "name_len": Int
    },  # Signals being returned by the UDF, with the signal name and type.
)
def name_len(name):
    # This is to avoid a sleep statement in the tests, so that the end-to-end test
    # knows when UDF processing has started, since we are testing canceling
    # UDF processing.
    # This is done to emulate a user waiting for processing that is stuck,
    # and pressing Ctrl-C to cancel the query script and UDF.
    print("UDF Processing Started")
    # Avoid any buffering so that the end-to-end test can react immediately.
    sys.stdout.flush()
    # Process very slowly to emulate a stuck script.
    sleep(1)
    if name.endswith(".json"):
        return (-1,)
    return (len(name),)


# Save as a new dataset.
DatasetQuery(
    path="gcs://dvcx-datalakes/dogs-and-cats/",
    anon=True,
).filter(C.name.glob("*cat*")).add_signals(name_len, parallel=1).save("name_len")  # type: ignore[attr-defined]



import io

import pytest

from dvcx.client.fileslice import FileSlice


def test_positions():
    data = b"0123456789abcdef"
    base = io.BytesIO(data)
    f = FileSlice(base, 5, 5, "foo")
    assert base.tell() == 0
    assert f.readable()
    assert not f.writable()
    assert f.seekable()
    assert f.name == "foo"

    # f.seek() doesn't move the underlying stream
    f.seek(0)
    assert f.tell() == 0
    assert base.tell() == 0

    assert f.read(3) == data[5:8]
    assert f.tell() == 3
    assert base.tell() == 8

    assert f.read(4) == data[8:10]
    assert f.tell() == 5
    assert base.tell() == 10

    b = bytearray(5)
    f.seek(0)
    f.readinto(b)
    assert b == data[5:10]


def test_invalid_slice():
    data = b"0123456789abcdef"
    base = io.BytesIO(data)
    f = FileSlice(base, 10, 10, "foo")
    assert f.read(4) == data[10:14]
    with pytest.raises(RuntimeError):
        f.read()


def test_close():
    data = b"0123456789abcdef"
    base = io.BytesIO(data)
    f = FileSlice(base, 5, 5, "foo")
    assert f.closed is False
    with f:
        assert f.closed is False
    assert f.closed is True
    assert base.closed is True


def test_implicit_close():
    # Assumes refcounting semantics
    data = b"0123456789abcdef"
    base = io.BytesIO(data)
    f = FileSlice(base, 5, 5, "foo")
    assert base.closed is False
    f = None  # noqa: F841
    assert base.closed is True



import asyncio

import pytest
from fsspec.asyn import sync
from tqdm import tqdm

from dvcx.asyn import get_loop
from dvcx.client import Client
from tests.data import ENTRIES


@pytest.fixture
def client(cloud_server, cloud_server_credentials):
    uri = cloud_server.src_uri
    return Client.get_implementation(uri).from_source(
        uri, cache=None, **cloud_server.client_config
    )


def normalize_entries(entries):
    return {(e.parent, e.name) for e in entries}


def match_entries(result, expected):
    assert len(result) == len(expected)
    assert normalize_entries(result) == normalize_entries(expected)


async def find(client, prefix):
    results = []
    try:
        async for entries in client.scandir(prefix):
            results.extend(entries)
    except BaseException as exc:
        print(f"find: {exc!r}")
        raise
    return results


def scandir(client, prefix):
    return sync(get_loop(), find, client, prefix)


def test_scandir_error(client):
    with pytest.raises(FileNotFoundError):
        scandir(client, "bogus")


@pytest.mark.xfail
def test_scandir_not_dir(client):
    with pytest.raises(FileNotFoundError):
        scandir(client, "description")


def test_scandir_success(client):
    results = scandir(client, "")
    match_entries(results, ENTRIES)


def test_gcs_client_gets_credentials_from_env(monkeypatch, mocker):
    from dvcx.client.gcs import GCSClient

    monkeypatch.setenv("DVCX_GCP_CREDENTIALS", '{"token": "test-credentials-token"}')
    init = mocker.patch("dvcx.client.gcs.GCSFileSystem.__init__", return_value=None)
    mocker.patch("dvcx.client.gcs.GCSFileSystem.invalidate_cache", return_value=None)

    GCSClient.create_fs()

    init.assert_called_once_with(
        token={"token": "test-credentials-token"}, version_aware=True
    )


@pytest.mark.parametrize("tree", [{}], indirect=True)
def test_fetch_dir_does_not_return_self(client, cloud_type):
    if cloud_type == "file":
        pytest.skip()

    client.fs.touch(f"{client.uri}/directory//file")

    subdirs = sync(
        get_loop(), client._fetch_dir, "directory/", tqdm(disable=True), asyncio.Queue()
    )

    assert "directory" not in subdirs



import inspect
from io import BytesIO
from typing import Any, Callable, Optional

try:
    from PIL import Image
except ImportError as exc:
    raise ImportError(
        "Missing dependencies for computer vision:\n"
        "To install run:\n\n"
        "  pip install 'dvcx[cv]'\n"
    ) from exc

from dvcx.lib.file import BinaryFile
from dvcx.lib.reader import FeatureReader


def convert_image(
    img: Image.Image,
    mode: str = "RGB",
    size: Optional[tuple[int, int]] = None,
    transform: Optional[Callable] = None,
    open_clip_model: Optional[Any] = None,
):
    """
    Resize, transform, and otherwise convert an image.

    Args:
        img (Image): PIL.Image object.
        mode (str): PIL.Image mode.
        size (tuple[int, int]): Size in (width, height) pixels for resizing.
        transform (Callable): Torchvision v1 or other transform to apply.
        open_clip_model (Any): Encode image using model from open_clip library.
    """
    if mode:
        img = img.convert(mode)
    if size:
        img = img.resize(size)
    if transform:
        img = transform(img)
        if open_clip_model:
            img = img.unsqueeze(0)  # type: ignore[attr-defined]
    if open_clip_model:
        method_name = "encode_image"
        if not (
            hasattr(open_clip_model, method_name)
            and inspect.ismethod(getattr(open_clip_model, method_name))
        ):
            raise ValueError(
                f"Unable to render Image: 'open_clip_model' doesn't support"
                f" '{method_name}()'"
            )
        img = open_clip_model.encode_image(img)
    return img


class ImageReader(FeatureReader):
    def __init__(
        self,
        mode: str = "RGB",
        size: Optional[tuple[int, int]] = None,
        transform: Optional[Callable] = None,
        open_clip_model: Any = None,
    ):
        """
        Read and optionally transform an image.

        All kwargs are passed to `convert_image()`.
        """
        self.mode = mode
        self.size = size
        self.transform = transform
        self.open_clip_model = open_clip_model
        super().__init__(BinaryFile)

    def __call__(self, value: bytes):
        img = Image.open(BytesIO(value))
        return convert_image(
            img,
            mode=self.mode,
            size=self.size,
            transform=self.transform,
            open_clip_model=self.open_clip_model,
        )



import asyncio
from collections.abc import Awaitable, Coroutine, Iterable
from concurrent.futures import ThreadPoolExecutor
from heapq import heappop, heappush
from typing import (
    Any,
    Callable,
    Generic,
    Optional,
    TypeVar,
)

from fsspec.asyn import get_loop

ASYNC_WORKERS = 20

InputT = TypeVar("InputT", contravariant=True)  # noqa: PLC0105
ResultT = TypeVar("ResultT", covariant=True)  # noqa: PLC0105


class AsyncMapper(Generic[InputT, ResultT]):
    """
    Asynchronous unordered mapping iterable compatible with fsspec.

    `AsyncMapper(func, it)` is roughly equivalent to `map(func, it)`, except
    that `func` is an async function, which is executed concurrently by up to
    `workers` asyncio coroutines, and the results are yielded in arbitrary
    order.

    If `func` needs to call synchronous functions that may themselves run coroutines
    on the same loop, it must use `mapper.to_thread()`.

    Note that `loop`, which defaults to the fsspec loop, must be running on a different
    thread than the one calling `mapper.iterate()`.
    """

    order_preserving = False
    work_queue: "asyncio.Queue[InputT]"
    loop: asyncio.AbstractEventLoop

    def __init__(
        self,
        func: Callable[[InputT], Awaitable[ResultT]],
        iterable: Iterable[InputT],
        *,
        workers: int = ASYNC_WORKERS,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ):
        self.func = func
        self.iterable = iterable
        self.workers = workers
        self.loop = get_loop() if loop is None else loop
        self.pool = ThreadPoolExecutor(workers)
        self._tasks: set[asyncio.Task] = set()

    def start_task(self, coro: Coroutine) -> asyncio.Task:
        task = self.loop.create_task(coro)
        self._tasks.add(task)
        task.add_done_callback(self._tasks.discard)
        return task

    async def produce(self) -> None:
        for item in self.iterable:
            await self.work_queue.put(item)

    async def worker(self) -> None:
        while (item := await self.work_queue.get()) is not None:
            try:
                result = await self.func(item)
                await self.result_queue.put(result)
            finally:
                self.work_queue.task_done()

    async def init(self) -> None:
        self.work_queue = asyncio.Queue(2 * self.workers)
        self.result_queue: asyncio.Queue[Optional[ResultT]] = asyncio.Queue(
            self.workers
        )

    async def run(self) -> None:
        producer = self.start_task(self.produce())
        for _i in range(self.workers):
            self.start_task(self.worker())
        try:
            done, pending = await asyncio.wait(
                self._tasks, return_when=asyncio.FIRST_COMPLETED
            )
            self.gather_exceptions(done)
            assert producer.done()
            join = self.start_task(self.work_queue.join())
            done, pending = await asyncio.wait(
                self._tasks, return_when=asyncio.FIRST_COMPLETED
            )
            self.gather_exceptions(done)
            assert join.done()
        except:
            await self.cancel_all()
            await self._break_iteration()
            raise
        else:
            await self.cancel_all()
            await self._end_iteration()

    async def cancel_all(self) -> None:
        if self._tasks:
            for task in self._tasks:
                task.cancel()
            await asyncio.wait(self._tasks)

    def gather_exceptions(self, done_tasks):
        # Check all exceptions to avoid "Task exception was never retrieved" warning
        exceptions = [task.exception() for task in done_tasks]
        # Raise the first exception found, if any. Additional ones are ignored.
        for exc in exceptions:
            if exc:
                raise exc

    async def _pop_result(self) -> Optional[ResultT]:
        return await self.result_queue.get()

    def next_result(self, timeout=None) -> Optional[ResultT]:
        """
        Return the next available result.

        Blocks as long as the result queue is empty.
        """
        future = asyncio.run_coroutine_threadsafe(self._pop_result(), self.loop)
        return future.result(timeout=timeout)

    async def _end_iteration(self) -> None:
        """Signal successful end of iteration."""
        await self.result_queue.put(None)

    async def _break_iteration(self) -> None:
        """Signal that iteration must stop ASAP."""
        while not self.result_queue.empty():
            self.result_queue.get_nowait()
        await self.result_queue.put(None)

    def iterate(self, timeout=None) -> Iterable[ResultT]:
        init = asyncio.run_coroutine_threadsafe(self.init(), self.loop)
        init.result(timeout=1)
        async_run = asyncio.run_coroutine_threadsafe(self.run(), self.loop)
        try:
            while True:
                if (result := self.next_result(timeout)) is not None:
                    yield result
                else:
                    break
            if exc := async_run.exception():
                raise exc
        finally:
            if not async_run.done():
                async_run.cancel()

    def __iter__(self):
        return self.iterate()

    async def to_thread(self, func, *args):
        return await self.loop.run_in_executor(self.pool, func, *args)


class OrderedMapper(AsyncMapper[InputT, ResultT]):
    """
    Asynchronous ordered mapping iterable compatible with fsspec.

    See `AsyncMapper` for details.
    """

    order_preserving = True

    def __init__(
        self,
        func: Callable[[InputT], Awaitable[ResultT]],
        iterable: Iterable[InputT],
        *,
        workers: int = ASYNC_WORKERS,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ):
        super().__init__(func, iterable, workers=workers, loop=loop)
        self._waiters: dict[int, Any] = {}
        self._getters: dict[int, asyncio.Future[Optional[ResultT]]] = {}
        self.heap: list[tuple[int, Optional[ResultT]]] = []
        self._next_yield = 0
        self._items_seen = 0
        self._window = 2 * workers

    def _push_result(self, i: int, result: Optional[ResultT]) -> None:
        if i in self._getters:
            future = self._getters.pop(i)
            future.set_result(result)
        else:
            heappush(self.heap, (i, result))

    async def worker(self) -> None:
        while (item := await self.work_queue.get()) is not None:
            i = self._items_seen
            self._items_seen += 1
            if i >= self._next_yield + self._window:
                event = self._waiters[i - self._window] = asyncio.Event()
                await event.wait()
            result = await self.func(item)
            self._push_result(i, result)
            self.work_queue.task_done()

    async def init(self) -> None:
        self.work_queue = asyncio.Queue(2 * self.workers)

    async def _pop_result(self) -> Optional[ResultT]:
        if self.heap and self.heap[0][0] == self._next_yield:
            i, out = heappop(self.heap)
        else:
            self._getters[self._next_yield] = get_value = self.loop.create_future()
            out = await get_value
        if self._next_yield in self._waiters:
            event = self._waiters.pop(self._next_yield)
            event.set()
        self._next_yield += 1
        return out

    async def _end_iteration(self) -> None:
        self._push_result(self._next_yield + len(self.heap), None)

    async def _break_iteration(self) -> None:
        self.heap = []
        self._push_result(self._next_yield, None)



import base64
import pickle

import pytest
from sqlalchemy import Column, Integer, Table

from dvcx.data_storage.serializer import deserialize
from dvcx.data_storage.sqlite import SQLiteDatabaseEngine


@pytest.mark.parametrize("db_file", [":memory:", "file.db"])
def test_init_clone(db_file):
    db = SQLiteDatabaseEngine.from_db_file(db_file)
    assert db.db_file == db_file

    # Test clone
    db2 = db.clone()
    assert isinstance(db2, SQLiteDatabaseEngine)
    assert db2.db_file == db_file


def test_serialize():
    obj = SQLiteDatabaseEngine.from_db_file(":memory:")

    # Test serialization
    serialized = obj.serialize()
    assert serialized
    serialized_pickled = base64.b64decode(serialized.encode())
    assert serialized_pickled
    (f, args, kwargs) = pickle.loads(serialized_pickled)  # noqa: S301
    assert f == SQLiteDatabaseEngine.from_db_file
    assert args == [":memory:"]
    assert kwargs == {}

    # Test deserialization
    obj3 = deserialize(serialized)
    assert isinstance(obj3, SQLiteDatabaseEngine)
    assert obj3.db_file == ":memory:"
    assert obj3.clone_params() == obj.clone_params()


def test_table(sqlite_db):
    table = Table(
        "test_table", sqlite_db.metadata, Column("id", Integer, primary_key=True)
    )
    assert not sqlite_db.has_table("test_table")
    assert not sqlite_db.has_table("test_table_2")

    table.create(sqlite_db.engine)
    assert sqlite_db.has_table("test_table")
    assert not sqlite_db.has_table("test_table_2")

    sqlite_db.rename_table("test_table", "test_table_2")
    assert sqlite_db.has_table("test_table_2")
    assert not sqlite_db.has_table("test_table")

    sqlite_db.drop_table(Table("test_table_2", sqlite_db.metadata))
    assert not sqlite_db.has_table("test_table")
    assert not sqlite_db.has_table("test_table_2")


def test_table_in_transaction(sqlite_db):
    table = Table(
        "test_table", sqlite_db.metadata, Column("id", Integer, primary_key=True)
    )
    assert not sqlite_db.has_table("test_table")
    assert not sqlite_db.has_table("test_table_2")

    with sqlite_db.transaction():
        table.create(sqlite_db.engine)
        assert sqlite_db.has_table("test_table")
        assert not sqlite_db.has_table("test_table_2")

        sqlite_db.rename_table("test_table", "test_table_2")
        assert sqlite_db.has_table("test_table_2")
        assert not sqlite_db.has_table("test_table")

        sqlite_db.drop_table(Table("test_table_2", sqlite_db.metadata))
        assert not sqlite_db.has_table("test_table")
        assert not sqlite_db.has_table("test_table_2")



import os

import pandas as pd

from dvcx.lib.claude import (
    Rating,
    TextAnalytics,
)
from dvcx.lib.dataset import Dataset
from dvcx.lib.file import File
from dvcx.query.schema import C

SOURCE = "s3://ldb-public/remote/chatbots-public/"
model = "claude-3-opus-20240229"

PROMPT = """Consider the dialogue between the 'user' and the 'bot'. \
The 'user' is a human trying to find the best mobile plan. \
The 'bot' is a chatbot designed to query the user and offer the \
best  solution. The dialog is successful if the 'bot' is able to \
gather the information and offer a plan, or inform the user that \
such plan does not exist. The dialog is not successful if the \
conversation ends early or the 'user' requests additional functions \
the 'bot' cannot perform. Read the dialogue below and rate it 'Success' \
if it is successful, and 'Failure' if not. After that, provide \
one-sentence explanation of the reasons for this rating. Use only \
JSON object as output with the keys 'status', and 'explanation'.
"""

chats = Dataset(SOURCE, anon=True).filter(C.name.glob("*.txt")).limit(5)

content = chats.map(
    TextAnalytics(
        prompt=PROMPT,
        model_name=model,
        temperature=0.9,
        api_key=os.getenv("ANTHROPIC_API_KEY"),
    ),
    parallel=3,
)

with pd.option_context("display.max_columns", None):
    df = content.select(File.name, Rating).to_pandas()
    print(df)

    errors = df[df.rating__error != ""]
    if len(errors) > 0:
        print("Errors: ", len(errors))



import re

import pytest

from dvcx.error import DatasetNotFoundError
from dvcx.query import DatasetQuery, Session


def test_ephemeral_dataset_naming(catalog):
    session_name = "qwer45"

    with pytest.raises(ValueError):
        Session("wrong-ds_name", catalog=catalog)

    with Session(session_name, catalog=catalog) as session:
        ds_name = "my_test_ds12"
        session.catalog.create_dataset(ds_name)
        ds_tmp = DatasetQuery(
            name=ds_name, session=session, catalog=session.catalog
        ).save()
        session_uuid = f"[0-9a-fA-F]{{{Session.SESSION_UUID_LEN}}}"
        table_uuid = f"[0-9a-fA-F]{{{Session.TEMP_TABLE_UUID_LEN}}}"

        name_prefix = f"{Session.DATASET_PREFIX}{session_name}"
        pattern = rf"^{name_prefix}_{session_uuid}_{table_uuid}$"

        assert re.match(pattern, ds_tmp.name) is not None


def test_global_session_naming(catalog):
    session_uuid = f"[0-9a-fA-F]{{{Session.SESSION_UUID_LEN}}}"
    table_uuid = f"[0-9a-fA-F]{{{Session.TEMP_TABLE_UUID_LEN}}}"

    ds_name = "qwsd"
    catalog.create_dataset(ds_name)
    ds_tmp = DatasetQuery(name=ds_name, catalog=catalog).save()
    global_prefix = f"{Session.DATASET_PREFIX}{Session.GLOBAL_SESSION_NAME}"
    pattern = rf"^{global_prefix}_{session_uuid}_{table_uuid}$"
    assert re.match(pattern, ds_tmp.name) is not None


def test_ephemeral_dataset_lifecycle(catalog):
    session_name = "asd3d4"
    with Session(session_name, catalog=catalog) as session:
        ds_name = "my_test_ds12"
        session.catalog.create_dataset(ds_name)
        ds_tmp = DatasetQuery(
            name=ds_name, session=session, catalog=session.catalog
        ).save()

        assert isinstance(ds_tmp, DatasetQuery)
        assert ds_tmp.name != ds_name
        assert ds_tmp.name.startswith(Session.DATASET_PREFIX)
        assert session_name in ds_tmp.name

        ds = catalog.get_dataset(ds_tmp.name)
        assert ds is not None

    with pytest.raises(DatasetNotFoundError):
        catalog.get_dataset(ds_tmp.name)



"""
To install dependencies:

  pip install imgbeddings

"""

import uuid

from imgbeddings import imgbeddings

from dvcx.lib.param import Image
from dvcx.query import C, DatasetQuery, udf
from dvcx.sql.types import Array, Float32


@udf(
    params=(Image(),),
    output={"embedding": Array(Float32)},
    method="embedding",
)
class ImageEmbeddings:
    def __init__(self):
        self.emb = imgbeddings()

    def embedding(self, img):
        emb = self.emb.to_embeddings(img)
        return (emb[0].tolist(),)


if __name__ == "__main__":
    ds_name = uuid.uuid4().hex
    print(f"Saving to dataset: {ds_name}")
    # Save as a new dataset
    DatasetQuery(path="gcs://dvcx-datalakes/dogs-and-cats/").filter(
        C.name.glob("*cat*.jpg")
    ).limit(5).add_signals(ImageEmbeddings).save(ds_name)

    for row in DatasetQuery(name=ds_name).results()[:2]:
        print("default columns: ", row[:-1])
        print("embedding[:10]:  ", row[-1][:10])
        print(f"type: {type(row[-1]).__name__}, len: {len(row[-1])}")
        print()



import os

from PIL import ImageFile, ImageFilter
from tabulate import tabulate

from dvcx.lib.image_transform import ImageTransform
from dvcx.query import C, DatasetQuery

cloud_prefix = "s3://"  # for GCP just switch to "gcs://"
bucket = "dvcx-50k-laion-files-writable"  # which bucket to use for both read and write
bucket_region = "us-east-2"  # no need to specify for GCP

file_type = "*.jpg"  # which files to use
blur_radius = 3  # how much to blur
filter_mod = 512  # how much of a subset of the data to use, i.e., 1/512

# only needed for AWS (no effect if using GCP)
os.environ["AWS_REGION"] = bucket_region
os.environ["AWS_DEFAULT_REGION"] = bucket_region

ImageFile.LOAD_TRUNCATED_IMAGES = True

# Custom filters can be implemented with the ImageFilter abstract class
# https://pillow.readthedocs.io/en/stable/reference/ImageFilter.html#PIL.ImageFilter.Filter
image_filter = ImageFilter.GaussianBlur(radius=blur_radius)

image_filter_udf = ImageTransform(
    image_filter=image_filter,
    bucket_name=bucket,
    prefix=cloud_prefix,
    output_folder="blur",
    file_prefix="blur_",
)

if __name__ == "__main__":
    data = (
        DatasetQuery(os.path.join(cloud_prefix, bucket))
        .filter(C.name.glob(file_type))
        .filter(C.random % filter_mod == 0)
        .generate(image_filter_udf)
        .results()
    )

    # Output the contents of the new dataset.
    print(tabulate(data))



import uuid

from zalando_splits_and_classes_ds import ds

from dvcx.query import C, DatasetQuery

ds_name = uuid.uuid4().hex
ds.save(ds_name)
print(ds_name)
for row in (
    DatasetQuery(name=ds_name)
    .select(C.source, C.parent, C.name, C("class"), C("label"))
    .order_by(C.random)
    .limit(8)
    .results()
):
    print(row)



import json
import os
from typing import Optional

params_cache: Optional[dict[str, str]] = None


def param(key: str, default: Optional[str] = None) -> Optional[str]:
    """Get query parameter."""
    if not isinstance(key, str):
        raise TypeError("Param key must be a string")

    global params_cache  # noqa: PLW0603
    if params_cache is None:
        if env_params := os.getenv("DVCX_QUERY_PARAMS"):
            try:
                params_parsed = json.loads(env_params)
            except (TypeError, ValueError):
                raise ValueError("Invalid params provided") from None
            if isinstance(params_parsed, dict):
                params_cache = params_parsed
            else:
                raise ValueError("Invalid params provided")
        else:
            params_cache = {}

    return params_cache.get(key, default)



from sqlalchemy.sql.functions import ReturnTypeFromArgs


class greatest(ReturnTypeFromArgs):  # noqa: N801
    inherit_cache = True


class least(ReturnTypeFromArgs):  # noqa: N801
    inherit_cache = True



import pandas as pd
import pytest

from dvcx.lib.file import File
from dvcx.lib.webdataset_meta import MergeParquetAndNpz, WDSMetaError


def test_rows_order():
    pq_name = "111.parquet"
    npz_name = "123.npz"

    args = [File(name=pq_name), File(name=npz_name)]

    npz, pq = MergeParquetAndNpz.get_meta_streams(args)
    assert pq.name == pq_name
    assert npz.name == npz_name

    args.reverse()
    npz, pq = MergeParquetAndNpz.get_meta_streams(args)
    assert pq.name == pq_name
    assert npz.name == npz_name


def test_rows_num():
    with pytest.raises(WDSMetaError):
        MergeParquetAndNpz.get_meta_streams([File(name="111.parquet")])

    with pytest.raises(WDSMetaError):
        MergeParquetAndNpz.get_meta_streams(
            [
                File(name="111.parquet"),
                File(name="111.npz"),
                File(name="1111.parquet"),
            ]
        )


def test_content():
    df = pd.DataFrame(
        {
            "uid": ["12345", "67890", "abcde", "f0123"],
            "text": ["28", "22", "we", "hello world"],
        }
    )
    npz = {
        "b32_img": [[3.14] * 2] * 4,
        "b32_txt": [[2.72] * 2] * 4,
        "l14_img": [[1.41] * 2] * 4,
        "l14_txt": [[6.28] * 2] * 4,
        "dedup": [[2.3] * 2] * 4,
    }

    name = "111.parquet"
    stream = File(name=name)
    objs = list(MergeParquetAndNpz.parse_metafiles(npz, df, stream))

    assert len(objs) == 4

    file, meta = objs[0]
    assert file.name == "0"
    assert file.parent == name

    assert meta.b32_img == npz["b32_img"][0]
    assert meta.l14_img == npz["l14_img"][0]



from .catalog import (
    QUERY_DATASET_PREFIX,
    QUERY_SCRIPT_CANCELED_EXIT_CODE,
    QUERY_SCRIPT_INVALID_LAST_STATEMENT_EXIT_CODE,
    Catalog,
    parse_edvcx_file,
)
from .loader import get_catalog

__all__ = [
    "Catalog",
    "get_catalog",
    "parse_edvcx_file",
    "QUERY_SCRIPT_INVALID_LAST_STATEMENT_EXIT_CODE",
    "QUERY_SCRIPT_CANCELED_EXIT_CODE",
    "QUERY_DATASET_PREFIX",
]



import pandas as pd

from dvcx.lib.dataset import C, Dataset
from dvcx.lib.file import File
from dvcx.lib.webdataset_meta import LaionMeta, parse_wds_meta

ds = Dataset("s3://dvcx-datacomp-small", anon=True)
ds = ds.filter(C.name.glob("0020f*"))
ds = ds.apply(parse_wds_meta)

ds = ds.select(
    File.name,
    File.parent,
    LaionMeta.uid,
    LaionMeta.original_width,
    LaionMeta.face_bboxes,
    LaionMeta.b32_img,
    LaionMeta.dedup,
)

with pd.option_context("display.max_columns", None):
    print(ds.to_pandas())



import json

import numpy as np
from scipy.spatial import distance


def euclidean_distance(a: str, b: str):
    a_np = np.array(json.loads(a))
    b_np = np.array(json.loads(b))

    return np.linalg.norm(b_np - a_np)


def cosine_distance(a: str, b: str):
    return distance.cosine(json.loads(a), json.loads(b))



import contextlib
from collections.abc import Iterator, Sequence
from itertools import chain
from multiprocessing import cpu_count
from queue import Empty, Full, Queue
from sys import stdin
from time import sleep
from types import GeneratorType
from typing import Any, Optional

import attrs
import multiprocess
from dill import load
from fsspec.callbacks import DEFAULT_CALLBACK, Callback
from multiprocess import get_context

from dvcx.catalog import Catalog
from dvcx.catalog.loader import get_distributed_class
from dvcx.query.batch import DatasetRowsBatch
from dvcx.query.dataset import (
    get_download_callback,
    get_generated_callback,
    get_processed_callback,
    process_udf_outputs,
)
from dvcx.query.udf import UDFBase, UDFFactory, UDFResult

DEFAULT_BATCH_SIZE = 10000
STOP_SIGNAL = "STOP"
OK_STATUS = "OK"
FINISHED_STATUS = "FINISHED"
FAILED_STATUS = "FAILED"
NOTIFY_STATUS = "NOTIFY"


def full_module_type_path(typ: type) -> str:
    return f"{typ.__module__}.{typ.__qualname__}"


def get_n_workers_from_arg(n_workers: Optional[int] = None) -> int:
    if not n_workers:
        return cpu_count()
    if n_workers < 1:
        raise RuntimeError("Must use at least one worker for parallel UDF execution!")
    return n_workers


# For more context on the get_from_queue and put_into_queue functions, see the
# discussion here:
# https://github.com/iterative/dvcx/pull/1297#issuecomment-2026308773
# This problem is not exactly described by, but is also related to these Python issues:
# https://github.com/python/cpython/issues/66587
# https://github.com/python/cpython/issues/88628
# https://github.com/python/cpython/issues/108645


def get_from_queue(queue: Queue) -> Any:
    """
    Gets an item from a queue.
    This is required to handle signals, such as KeyboardInterrupt exceptions
    while waiting for items to be available, although only on certain installations.
    (See the above comment for more context.)
    """
    while True:
        try:
            return queue.get_nowait()
        except Empty:
            sleep(0.01)


def put_into_queue(queue: Queue, item: Any) -> None:
    """
    Puts an item into a queue.
    This is required to handle signals, such as KeyboardInterrupt exceptions
    while waiting for items to be queued, although only on certain installations.
    (See the above comment for more context.)
    """
    while True:
        try:
            queue.put_nowait(item)
            return
        except Full:
            sleep(0.01)


def udf_entrypoint() -> int:
    # Load UDF info from stdin
    udf_info = load(stdin.buffer)  # noqa: S301

    (
        warehouse_class,
        warehouse_args,
        warehouse_kwargs,
    ) = udf_info["warehouse_clone_params"]
    warehouse = warehouse_class(*warehouse_args, **warehouse_kwargs)

    # Parallel processing (faster for more CPU-heavy UDFs)
    dispatch = UDFDispatcher(
        udf_info["udf"],
        udf_info["catalog_init"],
        udf_info["id_generator_clone_params"],
        udf_info["metastore_clone_params"],
        udf_info["warehouse_clone_params"],
        is_generator=udf_info.get("is_generator", False),
        cache=udf_info["cache"],
    )

    query = udf_info["query"]
    batching = udf_info["batching"]
    table = udf_info["table"]
    n_workers = udf_info["processes"]
    udf = udf_info["udf"]
    if n_workers is True:
        # Use default number of CPUs (cores)
        n_workers = None

    with contextlib.closing(
        batching(warehouse.dataset_select_paginated, query)
    ) as udf_inputs:
        download_cb = get_download_callback()
        processed_cb = get_processed_callback()
        generated_cb = get_generated_callback(dispatch.is_generator)
        try:
            udf_results = dispatch.run_udf_parallel(
                udf_inputs,
                n_workers=n_workers,
                processed_cb=processed_cb,
                download_cb=download_cb,
            )
            process_udf_outputs(warehouse, table, udf_results, udf, cb=generated_cb)
        finally:
            download_cb.close()
            processed_cb.close()
            generated_cb.close()

    warehouse.insert_rows_done(table)

    return 0


def udf_worker_entrypoint() -> int:
    return get_distributed_class().run_worker()


class UDFDispatcher:
    _batch_size: Optional[int] = None

    def __init__(
        self,
        udf,
        catalog_init_params,
        id_generator_clone_params,
        metastore_clone_params,
        warehouse_clone_params,
        cache,
        is_generator=False,
        buffer_size=DEFAULT_BATCH_SIZE,
    ):
        # isinstance cannot be used here, as dill packages the entire class definition,
        # and so these two types are not considered exactly equal,
        # even if they have the same import path.
        if full_module_type_path(type(udf)) != full_module_type_path(UDFFactory):
            self.udf = udf
        else:
            self.udf = None
            self.udf_factory = udf
        self.catalog_init_params = catalog_init_params
        (
            self.id_generator_class,
            self.id_generator_args,
            self.id_generator_kwargs,
        ) = id_generator_clone_params
        (
            self.metastore_class,
            self.metastore_args,
            self.metastore_kwargs,
        ) = metastore_clone_params
        (
            self.warehouse_class,
            self.warehouse_args,
            self.warehouse_kwargs,
        ) = warehouse_clone_params
        self.is_generator = is_generator
        self.cache = cache
        self.catalog = None
        self.task_queue = None
        self.done_queue = None
        self.buffer_size = buffer_size
        self.ctx = get_context("spawn")

    @property
    def batch_size(self):
        if not self.udf:
            self.udf = self.udf_factory()
        if self._batch_size is None:
            if hasattr(self.udf, "properties") and hasattr(
                self.udf.properties, "batch"
            ):
                self._batch_size = self.udf.properties.batch
            else:
                self._batch_size = 1
        return self._batch_size

    def _create_worker(self) -> "UDFWorker":
        if not self.catalog:
            id_generator = self.id_generator_class(
                *self.id_generator_args, **self.id_generator_kwargs
            )
            metastore = self.metastore_class(
                *self.metastore_args, **self.metastore_kwargs
            )
            warehouse = self.warehouse_class(
                *self.warehouse_args, **self.warehouse_kwargs
            )
            self.catalog = Catalog(
                id_generator, metastore, warehouse, **self.catalog_init_params
            )
        if not self.udf:
            self.udf = self.udf_factory()

        return UDFWorker(
            self.catalog,
            self.udf,
            self.task_queue,
            self.done_queue,
            self.is_generator,
            self.cache,
        )

    def _run_worker(self) -> None:
        try:
            worker = self._create_worker()
            worker.run()
        except (Exception, KeyboardInterrupt) as e:
            put_into_queue(self.done_queue, {"status": FAILED_STATUS, "exception": e})
            raise

    @staticmethod
    def send_stop_signal_to_workers(task_queue, n_workers: Optional[int] = None):
        n_workers = get_n_workers_from_arg(n_workers)
        for _ in range(n_workers):
            put_into_queue(task_queue, STOP_SIGNAL)

    def create_input_queue(self):
        return self.ctx.Queue()

    def run_udf_parallel(  # noqa: C901, PLR0912
        self,
        input_rows,
        n_workers: Optional[int] = None,
        cache: bool = False,
        input_queue=None,
        processed_cb: Callback = DEFAULT_CALLBACK,
        download_cb: Callback = DEFAULT_CALLBACK,
    ) -> Iterator[Sequence[UDFResult]]:
        n_workers = get_n_workers_from_arg(n_workers)

        if self.buffer_size < n_workers:
            raise RuntimeError(
                f"Parallel run error: buffer size is smaller than "
                f"number of workers: {self.buffer_size} < {n_workers}"
            )

        if input_queue:
            streaming_mode = True
            self.task_queue = input_queue
        else:
            streaming_mode = False
            self.task_queue = self.ctx.Queue()
        self.done_queue = self.ctx.Queue()
        pool = [
            self.ctx.Process(name=f"Worker-UDF-{i}", target=self._run_worker)
            for i in range(n_workers)
        ]
        for p in pool:
            p.start()

        # Will be set to True if all tasks complete normally
        normal_completion = False
        try:
            # Will be set to True when the input is exhausted
            input_finished = False

            if not streaming_mode:
                # Stop all workers after the input rows have finished processing
                input_data = chain(input_rows, [STOP_SIGNAL] * n_workers)

                # Add initial buffer of tasks
                for _ in range(self.buffer_size):
                    try:
                        put_into_queue(self.task_queue, next(input_data))
                    except StopIteration:
                        input_finished = True
                        break

            # Process all tasks
            while n_workers > 0:
                result = get_from_queue(self.done_queue)
                status = result["status"]
                if status == NOTIFY_STATUS:
                    download_cb.relative_update(result["downloaded"])
                elif status == FINISHED_STATUS:
                    # Worker finished
                    n_workers -= 1
                elif status == OK_STATUS:
                    processed_cb.relative_update(result["processed"])
                    yield result["result"]
                else:  # Failed / error
                    n_workers -= 1
                    exc = result.get("exception")
                    if exc:
                        raise exc
                    raise RuntimeError("Internal error: Parallel UDF execution failed")

                if not streaming_mode and not input_finished:
                    try:
                        put_into_queue(self.task_queue, next(input_data))
                    except StopIteration:
                        input_finished = True

            # Finished with all tasks normally
            normal_completion = True
        finally:
            if not normal_completion:
                # Stop all workers if there is an unexpected exception
                for _ in pool:
                    put_into_queue(self.task_queue, STOP_SIGNAL)
                self.task_queue.close()

                # This allows workers (and this process) to exit without
                # consuming any remaining data in the queues.
                # (If they exit due to an exception.)
                self.task_queue.cancel_join_thread()
                self.done_queue.cancel_join_thread()

                # Flush all items from the done queue.
                # This is needed if any workers are still running.
                while n_workers > 0:
                    result = get_from_queue(self.done_queue)
                    status = result["status"]
                    if status != OK_STATUS:
                        n_workers -= 1

            # Wait for workers to stop
            for p in pool:
                p.join()


class WorkerCallback(Callback):
    def __init__(self, queue: multiprocess.Queue):
        self.queue = queue

    def relative_update(self, inc: int = 1) -> None:
        put_into_queue(self.queue, {"status": NOTIFY_STATUS, "downloaded": inc})


@attrs.define
class UDFWorker:
    catalog: Catalog
    udf: UDFBase
    task_queue: multiprocess.Queue
    done_queue: multiprocess.Queue
    is_generator: bool
    cache: bool
    cb: Callback = attrs.field()

    @cb.default
    def _default_callback(self) -> WorkerCallback:
        return WorkerCallback(self.done_queue)

    def run(self) -> None:
        if hasattr(self.udf.func, "bootstrap") and callable(self.udf.func.bootstrap):
            self.udf.func.bootstrap()
        while (batch := get_from_queue(self.task_queue)) != STOP_SIGNAL:
            n_rows = len(batch.rows) if isinstance(batch, DatasetRowsBatch) else 1
            udf_output = self.udf(
                self.catalog,
                batch,
                is_generator=self.is_generator,
                cache=self.cache,
                cb=self.cb,
            )
            if isinstance(udf_output, GeneratorType):
                udf_output = list(udf_output)  # can not pickle generator
            put_into_queue(
                self.done_queue,
                {"status": OK_STATUS, "result": udf_output, "processed": n_rows},
            )

        if hasattr(self.udf.func, "teardown") and callable(self.udf.func.teardown):
            self.udf.func.teardown()

        put_into_queue(self.done_queue, {"status": FINISHED_STATUS})



import json
import os
from datetime import datetime
from typing import Any, cast

from dateutil.parser import isoparse
from gcsfs import GCSFileSystem

from dvcx.node import Entry

from .fsspec import DELIMITER, Client

# Patch gcsfs for consistency with s3fs
GCSFileSystem.set_session = GCSFileSystem._set_session


class GCSClient(Client):
    FS_CLASS = GCSFileSystem
    PREFIX = "gcs://"
    protocol = "gcs"

    @classmethod
    def create_fs(cls, **kwargs) -> GCSFileSystem:
        if os.environ.get("DVCX_GCP_CREDENTIALS"):
            kwargs["token"] = json.loads(os.environ["DVCX_GCP_CREDENTIALS"])
        if kwargs.pop("anon", False):
            kwargs["token"] = "anon"  # noqa: S105

        return cast(GCSFileSystem, super().create_fs(**kwargs))

    @staticmethod
    def parse_timestamp(timestamp: str) -> datetime:
        """
        Parse timestamp string returned by GCSFileSystem.

        This ensures that the passed timestamp is timezone aware.
        """
        dt = isoparse(timestamp)
        assert dt.tzinfo is not None
        return dt

    def convert_info(self, v: dict[str, Any], parent: str) -> Entry:
        name = v.get("name", "").split(DELIMITER)[-1]
        if "generation" in v:
            gen = f"#{v['generation']}"
            if name.endswith(gen):
                name = name[: -len(gen)]
        return Entry.from_file(
            parent=parent,
            name=name,
            etag=v.get("etag", ""),
            version=v.get("generation", ""),
            is_latest=not v.get("timeDeleted"),
            last_modified=self.parse_timestamp(v["updated"]),
            size=v.get("size", ""),
        )



import os
from unittest.mock import patch

import pytest


@pytest.mark.parametrize(
    "env_var,name,value",
    (
        ('{"foo":"bar"}', "foo", "bar"),
        ('{"foo":"bar2"}', "foo", "bar2"),
        ('{"foo":null}', "foo", None),
        ("{}", "foo", None),
        ("", "foo", None),
    ),
)
def test_query_param(env_var, name, value):
    with patch("dvcx.query.params.params_cache", None):
        with patch.dict(os.environ, {"DVCX_QUERY_PARAMS": env_var}):
            from dvcx.query.params import param

            assert param(name) == value


@pytest.mark.parametrize("env_var", ('"foo"', "[1,2,3]", "null", "1", "true", "foo"))
def test_query_param_error(env_var):
    with patch("dvcx.query.params.params_cache", None):
        with patch.dict(os.environ, {"DVCX_QUERY_PARAMS": env_var}):
            from dvcx.query.params import param

            with pytest.raises(ValueError):
                param("foo")


def test_query_param_key_error():
    from dvcx.query.params import param

    with pytest.raises(TypeError):
        param(12)



from datetime import datetime, timedelta, timezone

import pytest
from sqlalchemy import func

from dvcx import utils
from dvcx.error import StorageNotFoundError
from dvcx.storage import STALE_MINUTES_LIMIT, Status, Storage, StorageURI
from tests.utils import skip_if_not_sqlite

TS = datetime(2022, 8, 1)
EXPIRES = datetime(2022, 8, 2)


def test_human_time():
    assert utils.human_time_to_int("1236") == 1236
    assert utils.human_time_to_int("3h") == 3 * 60 * 60
    assert utils.human_time_to_int("2w") == 2 * 7 * 24 * 60 * 60
    assert utils.human_time_to_int("4M") == 4 * 31 * 24 * 60 * 60

    assert utils.human_time_to_int("bla") is None


def test_storage():
    s = Storage("s3://foo", TS, EXPIRES)

    d = s.to_dict()
    assert d.get("uri") == s.uri


def test_expiration_time():
    assert Storage.get_expiration_time(TS, 12344) == TS + timedelta(seconds=12344)


def test_adding_storage(metastore):
    uri = StorageURI("s3://whatever")
    with pytest.raises(StorageNotFoundError):
        metastore.get_storage(uri)

    storage, _, _, _, _ = metastore.register_storage_for_indexing(uri)
    cnt = next(metastore.db.execute(metastore._storages_select(func.count())), (0,))
    assert cnt[0] == 1

    bkt = list(
        metastore.db.execute(
            metastore._storages_select().where(metastore._storages.c.uri == uri)
        )
    )
    assert len(bkt) == 1

    assert storage.id == bkt[0][0]
    assert storage.uri == bkt[0][1]
    assert storage.timestamp == bkt[0][2]
    assert storage.expires == bkt[0][3]
    assert storage.started_inserting_at == bkt[0][4]
    assert storage.last_inserted_at == bkt[0][5]
    assert storage.status == bkt[0][6]
    assert storage.error_message == ""
    assert storage.error_stack == ""


def test_storage_status(metastore):
    uri = StorageURI("s3://somebucket")

    metastore.create_storage_if_not_registered(uri)
    storage = metastore.get_storage(uri)
    assert storage.uri == uri
    assert storage.status == Status.CREATED

    (
        storage,
        need_index,
        in_progress,
        partial_id,
        partial_path,
    ) = metastore.register_storage_for_indexing(uri)
    assert storage.status == Status.PENDING
    assert storage.uri == uri
    assert storage == metastore.get_storage(uri)
    assert need_index is True
    assert in_progress is False
    assert partial_id is None
    assert partial_path is None

    (
        s2,
        need_index,
        in_progress,
        partial_id,
        partial_path,
    ) = metastore.register_storage_for_indexing(uri)
    assert s2.status == Status.PENDING
    assert storage == s2 == metastore.get_storage(uri)
    assert need_index is False
    assert in_progress is True
    assert partial_id is None
    assert partial_path is None

    end_time = datetime.now(timezone.utc)
    metastore.mark_storage_indexed(uri, Status.COMPLETE, 1000, end_time)
    storage = metastore.get_storage(uri)
    assert storage.status == Status.COMPLETE


@pytest.mark.parametrize(
    "ttl",
    (-1, 999999999999, 99999999999999, 9999999999999999),
)
def test_max_ttl(ttl):
    uri = "s3://whatever"
    expires = Storage.get_expiration_time(TS, ttl)
    storage = Storage(1, uri, TS, expires)
    assert storage.timestamp == TS
    assert storage.expires == datetime.max
    assert storage.timestamp_str  # no error
    assert storage.timestamp_to_local  # no error
    assert storage.expires_to_local  # no error


def test_storage_without_dates():
    uri = "s3://whatever"
    storage = Storage(1, uri, None, None)
    assert storage.timestamp is None
    assert storage.expires is None
    assert storage.timestamp_str is None  # no error
    assert storage.timestamp_to_local is None  # no error
    assert storage.expires_to_local is None  # no error
    assert storage.to_dict() == {
        "uri": uri,
        "timestamp": None,
        "expires": None,
    }


async def test_storage_update_last_inserted_at(metastore):
    uri = StorageURI("s3://bucket_last_inserted")
    metastore.create_storage_if_not_registered(uri)
    metastore.update_last_inserted_at(uri)
    storage = metastore.get_storage(uri)
    assert storage.last_inserted_at


def test_stale_storage(metastore):
    uri_stale = StorageURI("s3://bucket_stale")
    uri_not_stale = StorageURI("s3://bucket_not_stale")

    metastore.create_storage_if_not_registered(uri_stale)
    metastore.create_storage_if_not_registered(uri_not_stale)

    metastore.mark_storage_pending(metastore.get_storage(uri_stale))
    metastore.mark_storage_pending(metastore.get_storage(uri_not_stale))

    # make storage looks stale
    updates = {
        "last_inserted_at": datetime.now(timezone.utc)
        - timedelta(minutes=STALE_MINUTES_LIMIT + 1)
    }
    s = metastore._storages
    metastore.db.execute(s.update().where(s.c.uri == uri_stale).values(**updates))

    metastore.find_stale_storages()

    stale_storage = metastore.get_storage(uri_stale)
    assert stale_storage.status == Status.STALE

    not_stale_storage = metastore.get_storage(uri_not_stale)
    assert not_stale_storage.status == Status.PENDING


def test_failed_storage(metastore):
    uri = StorageURI("s3://bucket")
    error_message = "Internal error on indexing"
    error_stack = "error"
    metastore.create_storage_if_not_registered(uri)

    metastore.mark_storage_pending(metastore.get_storage(uri))
    metastore.mark_storage_indexed(
        uri,
        Status.FAILED,
        1000,
        datetime.now(),
        error_message=error_message,
        error_stack=error_stack,
    )

    storage = metastore.get_storage(uri)
    assert storage.status == Status.FAILED
    assert storage.error_message == error_message
    assert storage.error_stack == error_stack


def test_unlist_source(
    listed_bucket,
    cloud_test_catalog,
    cloud_type,
):
    # TODO remove when https://github.com/iterative/dvcx/pull/868 is merged
    skip_if_not_sqlite()
    source_uri = cloud_test_catalog.src_uri
    catalog = cloud_test_catalog.catalog
    partial_id, partial_path = catalog.metastore.get_valid_partial_id(
        cloud_test_catalog.storage_uri, cloud_test_catalog.partial_path
    )
    storage_dataset_name = Storage.dataset_name(
        cloud_test_catalog.storage_uri, partial_path
    )

    # list source
    storage = catalog.get_storage(cloud_test_catalog.storage_uri)
    if cloud_type == "file":
        assert storage.status == Status.PARTIAL
    else:
        assert storage.status == Status.COMPLETE

    catalog.get_dataset(storage_dataset_name)

    # unlist source
    catalog.unlist_source(source_uri)
    with pytest.raises(StorageNotFoundError):
        catalog.get_storage(source_uri)
    # we preserve the table for dataset lineage
    catalog.get_dataset(storage_dataset_name)



"""
This module provides generic SQL functions for path logic.

These need to be implemented using dialect-specific compilation rules.
See https://docs.sqlalchemy.org/en/14/core/compiler.html
"""

from sqlalchemy.sql.functions import GenericFunction

from dvcx.sql.types import String
from dvcx.sql.utils import compiler_not_implemented


class parent(GenericFunction):  # noqa: N801
    """
    Returns the directory component of a posix-style path.
    """

    type = String()
    package = "path"
    name = "parent"
    inherit_cache = True


class name(GenericFunction):  # noqa: N801
    """
    Returns the final component of a posix-style path.
    """

    type = String()
    package = "path"
    name = "name"
    inherit_cache = True


class file_stem(GenericFunction):  # noqa: N801
    """
    Strips an extension from the given path.
    """

    type = String()
    package = "path"
    name = "file_stem"
    inherit_cache = True


class file_ext(GenericFunction):  # noqa: N801
    """
    Returns the extension of the given path.
    """

    type = String()
    package = "path"
    name = "file_ext"
    inherit_cache = True


compiler_not_implemented(parent)
compiler_not_implemented(name)
compiler_not_implemented(file_stem)
compiler_not_implemented(file_ext)



from dvcx.lib.iptc_exif_xmp import GetMetadata
from dvcx.query import C, DatasetQuery

source = "gcs://dvcx-datalakes/open-images-v6/"

if __name__ == "__main__":
    results = (
        DatasetQuery(source)
        .filter(C.name.glob("*.jpg"))
        .limit(10000)
        .add_signals(GetMetadata, parallel=True)
        .select("source", "xmp", "exif", "iptc", "error")
        .results()
    )
    print(*results, sep="\n")



import functools
import json
from abc import ABC, abstractmethod
from datetime import datetime, timezone
from fnmatch import fnmatch
from random import getrandbits
from typing import TYPE_CHECKING, Any, Callable, ClassVar, Optional, Union

import attrs
import sqlalchemy as sa
from fsspec.callbacks import DEFAULT_CALLBACK, Callback

from dvcx.data_storage.warehouse import RANDOM_BITS
from dvcx.sql.types import JSON, Boolean, DateTime, Int, SQLType, String

if TYPE_CHECKING:
    from dvcx.catalog import Catalog
    from dvcx.dataset import DatasetRow as Row


class ColumnMeta(type):
    def __getattr__(cls, name: str):
        return cls(name)


class Column(sa.ColumnClause, metaclass=ColumnMeta):
    inherit_cache: Optional[bool] = True

    def __init__(self, text, type_=None, is_literal=False, _selectable=None):
        self.name = text
        super().__init__(
            text, type_=type_, is_literal=is_literal, _selectable=_selectable
        )

    def glob(self, glob_str):
        return self.op("GLOB")(glob_str)


class UDFParameter(ABC):
    @abstractmethod
    def get_value(self, catalog: "Catalog", row: "Row", **kwargs) -> Any: ...

    async def get_value_async(
        self, catalog: "Catalog", row: "Row", mapper, **kwargs
    ) -> Any:
        return self.get_value(catalog, row, **kwargs)


@attrs.define(slots=False)
class ColumnParameter(UDFParameter):
    name: str

    def get_value(self, catalog, row, **kwargs):
        return row[self.name]


@attrs.define(slots=False)
class Object(UDFParameter):
    """
    Object is used as a placeholder parameter to indicate the actual stored object
    being passed as a parameter to the UDF.
    """

    reader: Callable

    def get_value(
        self,
        catalog: "Catalog",
        row: "Row",
        *,
        cache: bool = False,
        cb: Callback = DEFAULT_CALLBACK,
        **kwargs,
    ) -> Any:
        client = catalog.get_client(row.source)
        uid = row.as_uid()
        if cache:
            client.download(uid, callback=cb)
        with client.open_object(uid, use_cache=cache, cb=cb) as f:
            return self.reader(f)

    async def get_value_async(
        self,
        catalog: "Catalog",
        row: "Row",
        mapper,
        *,
        cache: bool = False,
        cb: Callback = DEFAULT_CALLBACK,
        **kwargs,
    ) -> Any:
        client = catalog.get_client(row.source)
        uid = row.as_uid()
        if cache:
            await client._download(uid, callback=cb)
        obj = await mapper.to_thread(
            functools.partial(client.open_object, uid, use_cache=cache, cb=cb)
        )
        with obj:
            return await mapper.to_thread(self.reader, obj)


@attrs.define(slots=False)
class Stream(UDFParameter):
    """
    A Stream() parameter receives a binary stream over the object contents.
    """

    def get_value(
        self,
        catalog: "Catalog",
        row: "Row",
        *,
        cache: bool = False,
        cb: Callback = DEFAULT_CALLBACK,
        **kwargs,
    ) -> Any:
        client = catalog.get_client(row.source)
        uid = row.as_uid()
        if cache:
            client.download(uid, callback=cb)
        return client.open_object(uid, use_cache=cache, cb=cb)

    async def get_value_async(
        self,
        catalog: "Catalog",
        row: "Row",
        mapper,
        *,
        cache: bool = False,
        cb: Callback = DEFAULT_CALLBACK,
        **kwargs,
    ) -> Any:
        client = catalog.get_client(row.source)
        uid = row.as_uid()
        if cache:
            await client._download(uid, callback=cb)
        return await mapper.to_thread(
            functools.partial(client.open_object, uid, use_cache=cache, cb=cb)
        )


@attrs.define(slots=False)
class LocalFilename(UDFParameter):
    """
    Placeholder parameter representing the local path to a cached copy of the object.

    If glob is None, then all files will be returned. If glob is specified,
    then only files matching the glob will be returned,
    otherwise None will be returned.
    """

    glob: Optional[str] = None

    def get_value(
        self,
        catalog: "Catalog",
        row: "Row",
        *,
        cb: Callback = DEFAULT_CALLBACK,
        **kwargs,
    ) -> Optional[str]:
        if self.glob and not fnmatch(row.name, self.glob):  # type: ignore[type-var]
            # If the glob pattern is specified and the row filename
            # does not match it, then return None
            return None
        client = catalog.get_client(row.source)
        uid = row.as_uid()
        client.download(uid, callback=cb)
        return client.cache.get_path(uid)

    async def get_value_async(
        self,
        catalog: "Catalog",
        row: "Row",
        mapper,
        *,
        cache: bool = False,
        cb: Callback = DEFAULT_CALLBACK,
        **kwargs,
    ) -> Optional[str]:
        if self.glob and not fnmatch(row.name, self.glob):  # type: ignore[type-var]
            # If the glob pattern is specified and the row filename
            # does not match it, then return None
            return None
        client = catalog.get_client(row.source)
        uid = row.as_uid()
        await client._download(uid, callback=cb)
        return client.cache.get_path(uid)


UDFParamSpec = Union[str, Column, UDFParameter]


def normalize_param(param: UDFParamSpec) -> UDFParameter:
    if isinstance(param, str):
        return ColumnParameter(param)
    if isinstance(param, Column):
        return ColumnParameter(param.name)
    if isinstance(param, UDFParameter):
        return param
    raise TypeError(f"Invalid UDF parameter: {param}")


class DatasetRow:
    schema: ClassVar[dict[str, type[SQLType]]] = {
        "source": String,
        "parent": String,
        "name": String,
        "size": Int,
        "location": JSON,
        "vtype": String,
        "dir_type": Int,
        "owner_name": String,
        "owner_id": String,
        "is_latest": Boolean,
        "last_modified": DateTime,
        "version": String,
        "etag": String,
        # system column
        "random": Int,
    }

    @staticmethod
    def create(
        name: str,
        source: str = "",
        parent: str = "",
        size: int = 0,
        location: Optional[dict[str, Any]] = None,
        vtype: str = "",
        dir_type: int = 0,
        owner_name: str = "",
        owner_id: str = "",
        is_latest: bool = True,
        last_modified: Optional[datetime] = None,
        version: str = "",
        etag: str = "",
    ) -> tuple[
        str,
        str,
        str,
        int,
        Optional[str],
        str,
        int,
        str,
        str,
        bool,
        datetime,
        str,
        str,
        int,
    ]:
        if location:
            location = json.dumps([location])  # type: ignore [assignment]

        last_modified = last_modified or datetime.now(timezone.utc)

        random = getrandbits(RANDOM_BITS)

        return (  # type: ignore [return-value]
            source,
            parent,
            name,
            size,
            location,
            vtype,
            dir_type,
            owner_name,
            owner_id,
            is_latest,
            last_modified,
            version,
            etag,
            random,
        )

    @staticmethod
    def extend(**columns):
        cols = {**DatasetRow.schema}
        cols.update(columns)
        return cols


C = Column



import importlib.util
import json
import os
import os.path as osp
import random
import stat
import sys
import time
from collections.abc import Iterable, Iterator, Sequence
from datetime import date, datetime, timezone
from itertools import islice
from typing import TYPE_CHECKING, Any, Optional, TypeVar, Union
from uuid import UUID

from dateutil import tz
from dateutil.parser import isoparse

if TYPE_CHECKING:
    import pandas as pd

GLOB_CHARS = ["?", "*", "[", "]"]
NUL = b"\0"
TIME_ZERO = datetime.fromtimestamp(0, tz=timezone.utc)

T = TypeVar("T", bound="DVCXDir")


class DVCXDir:
    DEFAULT = ".dvcx"
    CACHE = "cache"
    TMP = "tmp"
    DB = "db"
    ENV_VAR = "DVCX_DIR"
    ENV_VAR_DVCX_ROOT = "DVCX_ROOT_DIR"

    def __init__(
        self,
        root: Optional[str] = None,
        cache: Optional[str] = None,
        tmp: Optional[str] = None,
        db: Optional[str] = None,
    ) -> None:
        self.root = osp.abspath(root) if root is not None else self.default_root()
        self.cache = (
            osp.abspath(cache) if cache is not None else osp.join(self.root, self.CACHE)
        )
        self.tmp = (
            osp.abspath(tmp) if tmp is not None else osp.join(self.root, self.TMP)
        )
        self.db = osp.abspath(db) if db is not None else osp.join(self.root, self.DB)

    def init(self):
        os.makedirs(self.root, exist_ok=True)
        os.makedirs(self.cache, exist_ok=True)
        os.makedirs(self.tmp, exist_ok=True)
        os.makedirs(osp.split(self.db)[0], exist_ok=True)

    @classmethod
    def default_root(cls) -> str:
        try:
            root_dir = os.environ[cls.ENV_VAR_DVCX_ROOT]
        except KeyError:
            root_dir = os.getcwd()

        return osp.join(root_dir, cls.DEFAULT)

    @classmethod
    def find(cls: type[T], create: bool = True) -> T:
        try:
            root = os.environ[cls.ENV_VAR]
        except KeyError:
            root = cls.default_root()
        instance = cls(root)
        if not osp.isdir(root):
            if create:
                instance.init()
            else:
                NotADirectoryError(root)
        return instance


def human_time_to_int(time: str) -> Optional[int]:
    if not time:
        return None

    suffix = time[-1]
    try:
        num = int(time if suffix.isdigit() else time[:-1])
    except ValueError:
        return None
    return num * {
        "h": 60 * 60,
        "d": 60 * 60 * 24,
        "w": 60 * 60 * 24 * 7,
        "m": 31 * 24 * 60 * 60,
        "y": 60 * 60 * 24 * 365,
    }.get(suffix.lower(), 1)


def time_to_str(dt):
    if isinstance(dt, str):
        dt = isoparse(dt)
    return dt.strftime("%Y-%m-%d %H:%M:%S")


def time_to_local(dt: Union[datetime, str]) -> datetime:
    # TODO check usage
    if isinstance(dt, str):
        dt = isoparse(dt)
    try:
        return dt.astimezone(tz.tzlocal())
    except (OverflowError, OSError, ValueError):
        return dt


def time_to_local_str(dt: Union[datetime, str]) -> str:
    return time_to_str(time_to_local(dt))


def is_expired(expires: Optional[Union[datetime, str]]):
    if expires:
        return time_to_local(expires) < time_to_local(datetime.now())  # noqa: DTZ005

    return False


SIZE_SUFFIXES = ["", "K", "M", "G", "T", "P", "E", "Z", "Y", "R", "Q"]


def sizeof_fmt(num, suffix="", si=False):
    power = 1000.0 if si else 1024.0
    for unit in SIZE_SUFFIXES[:-1]:
        if abs(num) < power:
            if not unit:
                return f"{num:4.0f}{suffix}"
            return f"{num:3.1f}{unit}{suffix}"
        num /= power
    return f"{num:.1f}Q{suffix}"


def suffix_to_number(num_str: str) -> int:
    try:
        if len(num_str) > 1:
            suffix = num_str[-1].upper()
            if suffix in SIZE_SUFFIXES:
                suffix_idx = SIZE_SUFFIXES.index(suffix)
                return int(num_str[:-1]) * (1024**suffix_idx)
        return int(num_str)
    except (TypeError, ValueError):
        raise ValueError(f"Invalid number/suffix for: {num_str}") from None


def force_create_dir(name):
    if not os.path.exists(name):
        os.mkdir(name)
    elif not os.path.isdir(name):
        os.remove(name)
        os.mkdir(name)


def dvcx_paths_join(source_path: str, file_paths: Iterable[str]) -> Iterable[str]:
    source_parts = source_path.rstrip("/").split("/")
    if set(source_parts[-1]).intersection(GLOB_CHARS):
        # Remove last element if it is a glob match (such as *)
        source_parts.pop()
    source_stripped = "/".join(source_parts)
    return (f"{source_stripped}/{path.lstrip('/')}" for path in file_paths)


# From: https://docs.python.org/3/library/shutil.html#rmtree-example
def remove_readonly(func, path, _):
    "Clear the readonly bit and reattempt the removal"
    os.chmod(path, stat.S_IWRITE)
    func(path)


def sql_escape_like(search: str, escape: str = "\\") -> str:
    return (
        search.replace(escape, escape * 2)
        .replace("%", f"{escape}%")
        .replace("_", f"{escape}_")
    )


def get_envs_by_prefix(prefix: str) -> dict[str, str]:
    """
    Function that searches env variables by some name prefix and returns
    the ones found, but with prefix being excluded from it's names
    """
    variables: dict[str, str] = {}
    for env_name, env_value in os.environ.items():
        if env_name.startswith(prefix):
            variables[env_name[len(prefix) :]] = env_value

    return variables


def import_object(object_spec):
    filename, identifier = object_spec.rsplit(":", 1)
    filename = filename.strip()
    identifier = identifier.strip()

    if not identifier.isidentifier() or not filename.endswith(".py"):
        raise ValueError(f"Invalid object spec: {object_spec}")

    modname = os.path.abspath(filename)
    if modname in sys.modules:
        module = sys.modules[modname]
    else:
        # Use importlib to find and load the module from the given filename
        spec = importlib.util.spec_from_file_location(modname, filename)
        module = importlib.util.module_from_spec(spec)
        sys.modules[modname] = module
        spec.loader.exec_module(module)

    return getattr(module, identifier)


def parse_params_string(params: str):
    """
    Parse a string containing UDF class constructor parameters in the form
    `a, b, key=val` into *args and **kwargs.
    """
    args = []
    kwargs = {}
    for part in params.split():
        if "=" in part:
            key, val = part.split("=")
            kwargs[key] = val
        else:
            args.append(part)
    if any((args, kwargs)):
        return args, kwargs
    return None, None


_T_co = TypeVar("_T_co", covariant=True)


def batched(iterable: Iterable[_T_co], n: int) -> Iterator[tuple[_T_co, ...]]:
    "Batch data into tuples of length n. The last batch may be shorter."
    # Based on: https://docs.python.org/3/library/itertools.html#itertools-recipes
    # batched('ABCDEFG', 3) --> ABC DEF G
    if n < 1:
        raise ValueError("Batch size must be at least one")
    it = iter(iterable)
    while batch := tuple(islice(it, n)):
        yield batch


def flatten(items):
    for item in items:
        if isinstance(item, list):
            yield from item
        else:
            yield item


def retry_with_backoff(retries=5, backoff_sec=1):
    def retry(f):
        def wrapper(*args, **kwargs):
            num_tried = 0
            while True:
                try:
                    return f(*args, **kwargs)
                except Exception:
                    if num_tried == retries:
                        raise
                    sleep = (
                        backoff_sec * 2** num_tried + random.uniform(0, 1)  # noqa: S311
                    )
                    time.sleep(sleep)
                    num_tried += 1

        return wrapper

    return retry


def determine_processes(parallel: Optional[Union[bool, int]]) -> Union[bool, int]:
    if parallel is None and os.environ.get("DVCX_SETTINGS_PARALLEL") is not None:
        parallel = int(os.environ["DVCX_SETTINGS_PARALLEL"])
    if parallel is None or parallel is False:
        return False
    if parallel is True:
        return True
    if parallel == 0:
        return False
    if parallel < 0:
        return True
    return parallel


def get_env_list(
    key: str, default: Optional[Sequence] = None, sep: str = ","
) -> Optional[Sequence[str]]:
    try:
        str_val = os.environ[key]
    except KeyError:
        return default
    return str_val.split(sep=sep)


def show_df(
    df: "pd.DataFrame", collapse_columns: bool = True, system_columns: bool = False
) -> None:
    import pandas as pd

    if df.empty:
        return

    options: list[Any] = ["display.show_dimensions", False, "display.min_rows", 0]
    if not collapse_columns:
        options.extend(("display.max_columns", None))  # show all columns
        options.extend(("display.max_colwidth", None))  # do not truncate cells
        options.extend(("display.width", None))  #  do not truncate table

    if not system_columns:
        df.drop(
            columns=[
                "dir_type",
                "etag",
                "is_latest",
                "last_modified",
                "owner_id",
                "owner_name",
                "size",
                "version",
                "vtype",
            ],
            inplace=True,
            errors="ignore",
        )

    with pd.option_context("display.max_rows", None, *options):  # show all rows
        print(df)


def show_records(
    records: Optional[list[dict]],
    collapse_columns: bool = False,
    system_columns: bool = False,
) -> None:
    import pandas as pd

    if not records:
        return

    index = [row.pop("id") for row in records]
    df = pd.DataFrame.from_records(records, index=index)
    return show_df(df, collapse_columns=collapse_columns, system_columns=system_columns)


class JSONSerialize(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, bytes):
            return list(obj[:1024])
        if isinstance(obj, (datetime, date)):
            return obj.isoformat()
        if isinstance(obj, UUID):
            return str(obj)

        return super().default(obj)



import base64
import pickle

from dvcx.data_storage.serializer import deserialize
from dvcx.data_storage.sqlite import (
    SQLiteDatabaseEngine,
    SQLiteIDGenerator,
    SQLiteWarehouse,
)


def test_serialize():
    db = SQLiteDatabaseEngine.from_db_file(":memory:")
    id_generator = SQLiteIDGenerator(db, table_prefix="prefix")

    obj = SQLiteWarehouse(id_generator, db)
    assert obj.id_generator == id_generator
    assert obj.db == db

    # Test clone
    obj2 = obj.clone()
    assert isinstance(obj2, SQLiteWarehouse)
    assert obj2.id_generator.db.db_file == obj.id_generator.db.db_file
    assert obj2.id_generator._table_prefix == obj.id_generator._table_prefix
    assert obj2.db.db_file == db.db_file
    assert obj2.clone_params() == obj.clone_params()

    # Test serialization
    serialized = obj.serialize()
    assert serialized
    serialized_pickled = base64.b64decode(serialized.encode())
    assert serialized_pickled
    (f, args, kwargs) = pickle.loads(serialized_pickled)  # noqa: S301
    assert f == SQLiteWarehouse.init_after_clone
    assert args == []
    assert kwargs == {
        "id_generator_clone_params": id_generator.clone_params(),
        "db_clone_params": db.clone_params(),
    }

    # Test deserialization
    obj3 = deserialize(serialized)
    assert isinstance(obj3, SQLiteWarehouse)
    assert obj3.id_generator.db.db_file == id_generator.db.db_file
    assert obj3.id_generator._table_prefix == id_generator._table_prefix
    assert obj3.db.db_file == db.db_file
    assert obj3.clone_params() == obj.clone_params()


def test_is_temp_table_name(warehouse):
    assert warehouse.is_temp_table_name("tmp_vc12F") is True
    assert warehouse.is_temp_table_name("udf_jh653") is True
    assert warehouse.is_temp_table_name("ds_shadow_12345") is True
    assert warehouse.is_temp_table_name("old_ds_shadow") is True
    assert warehouse.is_temp_table_name("ds_my_dataset") is False
    assert warehouse.is_temp_table_name("src_my_bucket") is False
    assert warehouse.is_temp_table_name("ds_ds_my_query_script_1_1") is False


def test_dataset_stats_no_table(cloud_test_catalog, dogs_dataset):
    catalog = cloud_test_catalog.catalog
    catalog.warehouse.drop_dataset_rows_table(dogs_dataset, 1)
    num_objects, size = catalog.warehouse.dataset_stats(dogs_dataset, 1)
    assert num_objects is None
    assert size is None



from sqlalchemy.sql.expression import func

from . import path, string
from .conditional import greatest, least
from .random import rand

count = func.count
sum = func.sum
avg = func.avg
min = func.min
max = func.max

__all__ = [
    "avg",
    "count",
    "func",
    "greatest",
    "least",
    "max",
    "min",
    "path",
    "rand",
    "string",
    "sum",
]



import os

from langchain_community.embeddings import OpenAIEmbeddings
from scipy.spatial import distance

from dvcx.query import C, DatasetQuery, udf
from dvcx.sql.functions.array import cosine_distance
from dvcx.sql.types import Float

query = "What is the best way to do topic modeling?"

openai_api_key = os.environ["OPENAI_API_KEY"]
assert openai_api_key.startswith("sk-")
openai = OpenAIEmbeddings(openai_api_key=openai_api_key)
del openai_api_key

(embed_query,) = openai.embed_documents([query])


@udf(
    params=("embed",),
    output={"distance": Float},
)
def distance_to_example(embed, embed0=embed_query):
    dist = distance.cosine(embed, embed0)
    return (dist,)


DatasetQuery(name="pdf-bib").mutate(distance=cosine_distance(C.embed, embed_query))



# pip install Pillow torchvision

import torch
from torch import nn, optim
from torch.utils.data import DataLoader
from torchvision.transforms import v2

from dvcx.lib.dataset import C, Dataset
from dvcx.lib.image import ImageReader
from dvcx.lib.reader import LabelReader

STORAGE = "gcs://dvcx-datalakes/dogs-and-cats/"

# Define transformation for data preprocessing
transform = v2.Compose(
    [
        v2.ToTensor(),
        v2.Resize((64, 64)),
        v2.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
    ]
)


CLASSES = ["cat", "dog"]


# Define torch model
class CNN(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(3, 16, kernel_size=3, stride=2, padding=1)
        self.conv2 = nn.Conv2d(16, 32, kernel_size=3, stride=2, padding=1)
        self.conv3 = nn.Conv2d(32, 64, kernel_size=3, stride=2, padding=1)
        self.fc1 = nn.Linear(64 * 8 * 8, 512)
        self.fc2 = nn.Linear(512, len(CLASSES))

    def forward(self, x):
        x = torch.relu(self.conv1(x))
        x = torch.relu(self.conv2(x))
        x = torch.relu(self.conv3(x))
        x = x.view(-1, 64 * 8 * 8)
        x = torch.relu(self.fc1(x))
        return self.fc2(x)


if __name__ == "__main__":
    ds = (
        Dataset(STORAGE)
        .filter(C.name.glob("*.jpg"))
        .map(lambda name: (name[:3],), output={"label": str})
    )

    train_loader = DataLoader(
        ds.to_pytorch(
            ImageReader(),
            LabelReader(feature="label", classes=CLASSES),
            transform=transform,
        ),
        batch_size=16,
        num_workers=2,
    )

    model = CNN()
    criterion = nn.CrossEntropyLoss()
    optimizer = optim.Adam(model.parameters(), lr=0.001)

    # Train the model
    num_epochs = 10
    for epoch in range(num_epochs):
        for i, data in enumerate(train_loader):
            inputs, labels = data
            optimizer.zero_grad()

            # Forward pass
            outputs = model(inputs)
            loss = criterion(outputs, labels)

            # Backward pass and optimize
            loss.backward()
            optimizer.step()

            print("[%d, %5d] loss: %.3f" % (epoch + 1, i + 1, loss.item()))

    print("Finished Training")



from datetime import datetime
from typing import Any

import pytest

from dvcx.sql.types import (
    JSON,
    Array,
    Boolean,
    DateTime,
    Float,
    Float32,
    Float64,
    Int,
    String,
)
from tests.utils import DEFAULT_TREE, TARRED_TREE, create_tar_dataset

COMPLEX_TREE: dict[str, Any] = {
    **TARRED_TREE,
    **DEFAULT_TREE,
    "nested": {"dir": {"path": {"abc.txt": "abc"}}},
}


@pytest.mark.parametrize("tree", [COMPLEX_TREE], indirect=True)
def test_dir_expansion(cloud_test_catalog, version_aware, cloud_type):
    has_version = version_aware or cloud_type == "gcs"

    ctc = cloud_test_catalog
    catalog = ctc.catalog
    src_uri = ctc.src_uri
    if cloud_type == "file":
        # we don't want to index things in parent directory
        src_uri += "/"

    ds = create_tar_dataset(catalog, ctc.src_uri, "ds2")
    dataset = catalog.get_dataset(ds.name)
    st = catalog.warehouse.clone()
    q = st.dataset_rows(dataset).dir_expansion()
    columns = (
        "id",
        "vtype",
        "is_dir",
        "source",
        "parent",
        "name",
        "version",
        "location",
    )
    result = [dict(zip(columns, r)) for r in st.db.execute(q)]
    to_compare = [
        (r["parent"], r["name"], r["vtype"], r["is_dir"], r["version"] != "")
        for r in result
    ]

    assert all(r["source"] == ctc.storage_uri for r in result)
    if cloud_type == "file":
        prefix = ctc.partial_path + "/"
        prefix_root = ctc.partial_path
    else:
        prefix = ""
        prefix_root = ""

    # Note, we have both a file and a directory entry for expanded tar files
    expected = [
        (f"{prefix_root}", "animals.tar", "", 0, has_version),
        (f"{prefix_root}", "animals.tar", "", 1, False),
        (f"{prefix_root}", "cats", "", 1, False),
        (f"{prefix_root}", "description", "", 0, has_version),
        (f"{prefix_root}", "dogs", "", 1, False),
        (f"{prefix_root}", "nested", "", 1, False),
        (f"{prefix}animals.tar", "cats", "", 1, False),
        (f"{prefix}animals.tar", "description", "tar", 0, False),
        (f"{prefix}animals.tar", "dogs", "", 1, False),
        (f"{prefix}animals.tar/cats", "cat1", "tar", 0, False),
        (f"{prefix}animals.tar/cats", "cat2", "tar", 0, False),
        (f"{prefix}animals.tar/dogs", "dog1", "tar", 0, False),
        (f"{prefix}animals.tar/dogs", "dog2", "tar", 0, False),
        (f"{prefix}animals.tar/dogs", "dog3", "tar", 0, False),
        (f"{prefix}animals.tar/dogs", "others", "", 1, False),
        (f"{prefix}animals.tar/dogs/others", "dog4", "tar", 0, False),
        (f"{prefix}cats", "cat1", "", 0, has_version),
        (f"{prefix}cats", "cat2", "", 0, has_version),
        (f"{prefix}dogs", "dog1", "", 0, has_version),
        (f"{prefix}dogs", "dog2", "", 0, has_version),
        (f"{prefix}dogs", "dog3", "", 0, has_version),
        (f"{prefix}dogs", "others", "", 1, False),
        (f"{prefix}dogs/others", "dog4", "", 0, has_version),
        (f"{prefix}nested", "dir", "", 1, False),
        (f"{prefix}nested/dir", "path", "", 1, False),
        (f"{prefix}nested/dir/path", "abc.txt", "", 0, has_version),
    ]

    if cloud_type == "file":
        # since with file listing, parent is relative path to the root of FS as
        # storage uri is the root of FS, we need to add dirs to the root
        prefix_split = prefix.split("/")
        expected = [
            ("/".join(prefix_split[:i]), prefix_split[i], "", 1, False)
            for i in range(len(prefix_split) - 1)
        ] + expected

    assert to_compare == expected


@pytest.mark.parametrize(
    "cloud_type,version_aware",
    [("s3", True)],
    indirect=True,
)
def test_convert_type(cloud_test_catalog):
    ctc = cloud_test_catalog
    catalog = ctc.catalog
    warehouse = catalog.warehouse
    now = datetime.now()

    # convert int to float
    for f in [Float, Float32, Float64]:
        converted = warehouse.convert_type(1, f())
        assert converted == 1.0
        assert isinstance(converted, float)

    # types match, nothing to convert
    assert warehouse.convert_type(1, Int()) == 1
    assert warehouse.convert_type(1.5, Float()) == 1.5
    assert warehouse.convert_type(True, Boolean()) is True
    assert warehouse.convert_type("s", String()) == "s"
    assert warehouse.convert_type(now, DateTime()) == now
    assert warehouse.convert_type([1, 2], Array(Int)) == [1, 2]
    assert warehouse.convert_type([1.5, 2.5], Array(Float)) == [1.5, 2.5]
    assert warehouse.convert_type(["a", "b"], Array(String)) == ["a", "b"]
    assert warehouse.convert_type([[1, 2], [3, 4]], Array(Array(Int))) == [
        [1, 2],
        [3, 4],
    ]

    # JSON Tests
    assert warehouse.convert_type('{"a": 1}', JSON()) == '{"a": 1}'
    assert warehouse.convert_type({"a": 1}, JSON()) == '{"a": 1}'
    assert warehouse.convert_type([{"a": 1}], JSON()) == '[{"a": 1}]'
    with pytest.raises(ValueError):
        warehouse.convert_type(0.5, JSON())

    # convert array to compatible type
    converted = warehouse.convert_type([1, 2], Array(Float))
    assert converted == [1.0, 2.0]
    assert all(isinstance(c, float) for c in converted)

    # convert nested array to compatible type
    converted = warehouse.convert_type([[1, 2], [3, 4]], Array(Array(Float)))
    assert converted == [[1.0, 2.0], [3.0, 4.0]]
    assert all(isinstance(c, float) for c in converted[0])
    assert all(isinstance(c, float) for c in converted[1])

    # error, float to int
    with pytest.raises(ValueError):
        warehouse.convert_type(1.5, Int())

    # error, float to int in list
    with pytest.raises(ValueError):
        warehouse.convert_type([1.5, 1], Array(Int))



from .dataset import DatasetQuery
from .params import param
from .schema import C, DatasetRow, LocalFilename, Object, Stream
from .session import Session
from .udf import udf

__all__ = [
    "C",
    "DatasetQuery",
    "DatasetRow",
    "LocalFilename",
    "Object",
    "Session",
    "Stream",
    "param",
    "udf",
]



import logging
import os
import shlex
import sys
import traceback
from argparse import SUPPRESS, Action, ArgumentParser, ArgumentTypeError, Namespace
from collections.abc import Iterable, Iterator, Mapping, Sequence
from itertools import chain
from multiprocessing import freeze_support
from typing import TYPE_CHECKING, Optional, Union

import shtab

from dvcx import __version__, utils
from dvcx.cli_utils import BooleanOptionalAction, CommaSeparatedArgs, KeyValueArgs
from dvcx.utils import DVCXDir

if TYPE_CHECKING:
    from dvcx.catalog import Catalog

logger = logging.getLogger("dvcx")

TTL_HUMAN = "4h"
TTL_INT = 4 * 60 * 60
FIND_COLUMNS = ["du", "name", "owner", "path", "size", "type"]


def human_time_type(value_str: str, can_be_none: bool = False) -> Optional[int]:
    value = utils.human_time_to_int(value_str)

    if value:
        return value
    if can_be_none:
        return None

    raise ArgumentTypeError(
        "This option supports only a human-readable time interval like 12h or 4w."
    )


def parse_find_column(column: str) -> str:
    column_lower = column.strip().lower()
    if column_lower in FIND_COLUMNS:
        return column_lower
    raise ArgumentTypeError(
        f"Invalid column for find: '{column}' Options are: {','.join(FIND_COLUMNS)}"
    )


def find_columns_type(
    columns_str: str,
    default_colums_str: str = "path",
) -> list[str]:
    if not columns_str:
        columns_str = default_colums_str

    return [parse_find_column(c) for c in columns_str.split(",")]


def add_sources_arg(parser: ArgumentParser, nargs: Union[str, int] = "+") -> Action:
    return parser.add_argument(
        "sources",
        type=str,
        nargs=nargs,
        help="Data sources - paths to cloud storage dirs",
    )


def add_show_args(parser: ArgumentParser) -> None:
    parser.add_argument(
        "--limit",
        action="store",
        default=10,
        type=int,
        help="Number of rows to show",
    )
    parser.add_argument(
        "--offset",
        action="store",
        default=0,
        type=int,
        help="Number of rows to offset",
    )
    parser.add_argument(
        "--columns",
        default=[],
        action=CommaSeparatedArgs,
        help="Columns to show",
    )
    parser.add_argument(
        "--no-collapse",
        action="store_true",
        default=False,
        help="Do not collapse the columns",
    )


def get_parser() -> ArgumentParser:  # noqa: PLR0915
    parser = ArgumentParser(description="DVCx: Data Query Language", prog="dvcx")

    parser.add_argument("-V", "--version", action="version", version=__version__)
    parser.add_argument("--internal-run-udf", action="store_true", help=SUPPRESS)
    parser.add_argument("--internal-run-udf-worker", action="store_true", help=SUPPRESS)

    parent_parser = ArgumentParser(add_help=False)
    parent_parser.add_argument(
        "--aws-endpoint-url",
        type=str,
        help="AWS endpoint URL",
    )
    parent_parser.add_argument(
        "--anon",
        action="store_true",
        help="AWS anon (aka awscli's --no-sign-request)",
    )
    parent_parser.add_argument(
        "--ttl",
        type=human_time_type,
        default=TTL_HUMAN,
        help="Time-to-live of data source cache. Negative equals forever.",
    )
    parent_parser.add_argument(
        "-u", "--update", action="count", default=0, help="Update cache"
    )
    parent_parser.add_argument(
        "-v", "--verbose", action="count", default=0, help="Verbose"
    )
    parent_parser.add_argument(
        "-q", "--quiet", action="count", default=0, help="Be quiet"
    )
    parent_parser.add_argument(
        "--debug-sql",
        action="store_true",
        default=False,
        help="Show All SQL Queries (very verbose output, for debugging only)",
    )
    parent_parser.add_argument(
        "--pdb",
        action="store_true",
        default=False,
        help="Drop into the pdb debugger on fatal exception",
    )

    subp = parser.add_subparsers(help="Sub-command help", dest="command")
    parse_cp = subp.add_parser(
        "cp", parents=[parent_parser], help="Copy data files from the cloud"
    )
    add_sources_arg(parse_cp).complete = shtab.DIR  # type: ignore[attr-defined]
    parse_cp.add_argument("output", type=str, help="Output")
    parse_cp.add_argument(
        "-f",
        "--force",
        default=False,
        action="store_true",
        help="Force creating outputs",
    )
    parse_cp.add_argument(
        "-r",
        "-R",
        "--recursive",
        default=False,
        action="store_true",
        help="Copy directories recursively",
    )
    parse_cp.add_argument(
        "--no-glob",
        default=False,
        action="store_true",
        help="Do not expand globs (such as * or ?)",
    )

    parse_clone = subp.add_parser(
        "clone", parents=[parent_parser], help="Copy data files from the cloud"
    )
    add_sources_arg(parse_clone).complete = shtab.DIR  # type: ignore[attr-defined]
    parse_clone.add_argument("output", type=str, help="Output")
    parse_clone.add_argument(
        "-f",
        "--force",
        default=False,
        action="store_true",
        help="Force creating outputs",
    )
    parse_clone.add_argument(
        "-r",
        "-R",
        "--recursive",
        default=False,
        action="store_true",
        help="Copy directories recursively",
    )
    parse_clone.add_argument(
        "--no-glob",
        default=False,
        action="store_true",
        help="Do not expand globs (such as * or ?)",
    )
    parse_clone.add_argument(
        "--no-cp",
        default=False,
        action="store_true",
        help="Do not copy files, just create a dataset",
    )
    parse_clone.add_argument(
        "--edvcx",
        default=False,
        action="store_true",
        help="Create a .edvcx file",
    )
    parse_clone.add_argument(
        "--edvcx-file",
        help="Use a different filename for the resulting .edvcx file",
    )

    parse_pull = subp.add_parser(
        "pull", parents=[parent_parser], help="Pull specific dataset version from SaaS"
    )
    parse_pull.add_argument(
        "dataset",
        type=str,
        help="Name and version of remote dataset created in SaaS",
    )
    parse_pull.add_argument("-o", "--output", type=str, help="Output")
    parse_pull.add_argument(
        "-f",
        "--force",
        default=False,
        action="store_true",
        help="Force creating outputs",
    )
    parse_pull.add_argument(
        "-r",
        "-R",
        "--recursive",
        default=False,
        action="store_true",
        help="Copy directories recursively",
    )
    parse_pull.add_argument(
        "--no-cp",
        default=False,
        action="store_true",
        help="Do not copy files, just pull a remote dataset into local DB",
    )
    parse_pull.add_argument(
        "--edvcx",
        default=False,
        action="store_true",
        help="Create .edvcx file",
    )
    parse_pull.add_argument(
        "--edvcx-file",
        help="Use a different filename for the resulting .edvcx file",
    )

    parse_edit_dataset = subp.add_parser(
        "edit-dataset", parents=[parent_parser], help="Edit dataset metadata"
    )
    parse_edit_dataset.add_argument("name", type=str, help="Dataset name")
    parse_edit_dataset.add_argument(
        "--new-name",
        action="store",
        default="",
        help="Dataset new name",
    )
    parse_edit_dataset.add_argument(
        "--description",
        action="store",
        default="",
        help="Dataset description",
    )
    parse_edit_dataset.add_argument(
        "--labels",
        default=[],
        nargs="+",
        help="Dataset labels",
    )

    subp.add_parser("ls-datasets", parents=[parent_parser], help="List datasets")
    ls_dataset_rows_parser = subp.add_parser(
        "ls-dataset-rows", parents=[parent_parser], help="List dataset rows"
    )
    ls_dataset_rows_parser.add_argument("name", type=str, help="Dataset name")
    ls_dataset_rows_parser.add_argument(
        "--version",
        action="store",
        default=None,
        type=int,
        help="Dataset version",
    )

    rm_dataset_parser = subp.add_parser(
        "rm-dataset", parents=[parent_parser], help="Removes dataset"
    )
    rm_dataset_parser.add_argument("name", type=str, help="Dataset name")
    rm_dataset_parser.add_argument(
        "--version",
        action="store",
        default=None,
        type=int,
        help="Dataset version",
    )
    rm_dataset_parser.add_argument(
        "--force",
        default=False,
        action=BooleanOptionalAction,
        help="Force delete registered dataset with all of it's versions",
    )

    dataset_stats_parser = subp.add_parser(
        "dataset-stats", parents=[parent_parser], help="Shows basic dataset stats"
    )
    dataset_stats_parser.add_argument("name", type=str, help="Dataset name")
    dataset_stats_parser.add_argument(
        "--version",
        action="store",
        default=None,
        type=int,
        help="Dataset version",
    )
    dataset_stats_parser.add_argument(
        "-b",
        "--bytes",
        default=False,
        action="store_true",
        help="Display size in bytes instead of human-readable size",
    )
    dataset_stats_parser.add_argument(
        "--si",
        default=False,
        action="store_true",
        help="Display size using powers of 1000 not 1024",
    )

    parse_merge_datasets = subp.add_parser(
        "merge-datasets", parents=[parent_parser], help="Merges datasets"
    )
    parse_merge_datasets.add_argument(
        "--src",
        action="store",
        default=None,
        help="Source dataset name",
    )
    parse_merge_datasets.add_argument(
        "--dst",
        action="store",
        default=None,
        help="Destination dataset name",
    )
    parse_merge_datasets.add_argument(
        "--src-version",
        action="store",
        default=None,
        type=int,
        help="Source dataset version",
    )
    parse_merge_datasets.add_argument(
        "--dst-version",
        action="store",
        default=None,
        type=int,
        help="Destination dataset version",
    )

    parse_ls = subp.add_parser(
        "ls", parents=[parent_parser], help="List storage contents"
    )
    add_sources_arg(parse_ls, nargs="*")
    parse_ls.add_argument(
        "-l",
        "--long",
        action="count",
        default=0,
        help="List files in the long format",
    )
    parse_ls.add_argument(
        "--remote",
        action="store",
        default="",
        help="Name of remote to use",
    )

    parse_du = subp.add_parser(
        "du", parents=[parent_parser], help="Display space usage"
    )
    add_sources_arg(parse_du)
    parse_du.add_argument(
        "-b",
        "--bytes",
        default=False,
        action="store_true",
        help="Display sizes in bytes instead of human-readable sizes",
    )
    parse_du.add_argument(
        "-d",
        "--depth",
        "--max-depth",
        default=0,
        type=int,
        metavar="N",
        help=(
            "Display sizes for N directory depths below the given directory, "
            "the default is 0 (summarize provided directory only)."
        ),
    )
    parse_du.add_argument(
        "--si",
        default=False,
        action="store_true",
        help="Display sizes using powers of 1000 not 1024",
    )

    parse_find = subp.add_parser(
        "find", parents=[parent_parser], help="Search in a directory hierarchy"
    )
    add_sources_arg(parse_find)
    parse_find.add_argument(
        "--name",
        type=str,
        action="append",
        help="Filename to match pattern.",
    )
    parse_find.add_argument(
        "--iname",
        type=str,
        action="append",
        help="Like -name but case insensitive.",
    )
    parse_find.add_argument(
        "--path",
        type=str,
        action="append",
        help="Path to match pattern.",
    )
    parse_find.add_argument(
        "--ipath",
        type=str,
        action="append",
        help="Like -path but case insensitive.",
    )
    parse_find.add_argument(
        "--size",
        type=str,
        help=(
            "Filter by size (+ is greater or equal, - is less or equal). "
            "Specified size is in bytes, or use a suffix like K, M, G for "
            "kilobytes, megabytes, gigabytes, etc."
        ),
    )
    parse_find.add_argument(
        "--type",
        type=str,
        help='File type: "f" - regular, "d" - directory',
    )
    parse_find.add_argument(
        "-c",
        "--columns",
        type=find_columns_type,
        default=None,
        help=(
            "A comma-separated list of columns to print for each result. "
            f"Options are: {','.join(FIND_COLUMNS)} (Default: path)"
        ),
    )

    parse_index = subp.add_parser(
        "index", parents=[parent_parser], help="Index storage location"
    )
    add_sources_arg(parse_index)

    subp.add_parser(
        "find-stale-storages",
        parents=[parent_parser],
        help="Finds and marks stale storages",
    )

    show_parser = subp.add_parser(
        "show",
        parents=[parent_parser],
        help="Create a new dataset with a query script",
    )
    show_parser.add_argument("name", type=str, help="Dataset name")
    show_parser.add_argument(
        "--version",
        action="store",
        default=None,
        type=int,
        help="Dataset version",
    )
    add_show_args(show_parser)

    query_parser = subp.add_parser(
        "query",
        parents=[parent_parser],
        help="Create a new dataset with a query script",
    )
    query_parser.add_argument(
        "script", metavar="<script.py>", type=str, help="Filepath for script"
    )
    query_parser.add_argument(
        "dataset_name", nargs="?", type=str, help="Save result dataset as"
    )
    query_parser.add_argument(
        "--parallel",
        nargs="?",
        type=int,
        const=-1,
        default=None,
        metavar="N",
        help=(
            "Use multiprocessing to run any query script UDFs with N worker processes. "
            "N defaults to the CPU count."
        ),
    )
    add_show_args(query_parser)
    query_parser.add_argument(
        "-p",
        "--param",
        metavar="param=value",
        nargs=1,
        action=KeyValueArgs,
        help="Query parameters",
    )

    apply_udf_parser = subp.add_parser(
        "apply-udf", parents=[parent_parser], help="Apply UDF"
    )
    apply_udf_parser.add_argument("udf", type=str, help="UDF location")
    apply_udf_parser.add_argument("source", type=str, help="Source storage or dataset")
    apply_udf_parser.add_argument("target", type=str, help="Target dataset name")
    apply_udf_parser.add_argument(
        "--parallel",
        nargs="?",
        type=int,
        const=-1,
        default=None,
        metavar="N",
        help=(
            "Use multiprocessing to run the UDF with N worker processes. "
            "N defaults to the CPU count."
        ),
    )
    apply_udf_parser.add_argument(
        "--udf-params", type=str, default=None, help="UDF class parameters"
    )
    subp.add_parser(
        "clear-cache", parents=[parent_parser], help="Clear the local file cache"
    )
    subp.add_parser(
        "gc", parents=[parent_parser], help="Garbage collect temporary tables"
    )

    add_completion_parser(subp, [parent_parser])
    return parser


def add_completion_parser(subparsers, parents):
    parser = subparsers.add_parser(
        "completion",
        parents=parents,
        help="Output shell completion script",
    )
    parser.add_argument(
        "-s",
        "--shell",
        help="Shell syntax for completions.",
        default="bash",
        choices=shtab.SUPPORTED_SHELLS,
    )


def get_logging_level(args: Namespace) -> int:
    if args.quiet:
        return logging.CRITICAL
    if args.verbose:
        return logging.DEBUG
    return logging.INFO


def ls_urls(
    sources,
    catalog: "Catalog",
    long: bool = False,
    **kwargs,
) -> Iterator[tuple[str, Iterator[str]]]:
    curr_dir = None
    value_iterables = []
    for next_dir, values in _ls_urls_flat(sources, long, catalog, **kwargs):
        if curr_dir is None or next_dir == curr_dir:  # type: ignore[unreachable]
            value_iterables.append(values)
        else:
            yield curr_dir, chain(*value_iterables)  # type: ignore[unreachable]
            value_iterables = [values]
        curr_dir = next_dir
    if curr_dir is not None:
        yield curr_dir, chain(*value_iterables)


def _node_data_to_ls_values(row, long_format=False):
    from dvcx.node import DirType, long_line_str

    name = row[0]
    is_dir = row[1] == DirType.DIR
    ending = "/" if is_dir else ""
    value = name + ending
    if long_format:
        last_modified = row[2]
        owner_name = row[3]
        timestamp = last_modified if not is_dir else None
        return long_line_str(value, timestamp, owner_name)
    return value


def _ls_urls_flat(
    sources,
    long: bool,
    catalog: "Catalog",
    **kwargs,
) -> Iterator[tuple[str, Iterator[str]]]:
    from dvcx.client import Client
    from dvcx.node import long_line_str

    for source in sources:
        client_cls = Client.get_implementation(source)
        if client_cls.is_root_url(source):
            buckets = client_cls.ls_buckets(**catalog.client_config)
            if long:
                values = (long_line_str(b.name, b.created, "") for b in buckets)
            else:
                values = (b.name for b in buckets)
            yield source, values
        else:
            found = False
            fields = ["name", "dir_type"]
            if long:
                fields.extend(["last_modified", "owner_name"])
            for data_source, results in catalog.ls([source], fields=fields, **kwargs):
                values = (_node_data_to_ls_values(r, long) for r in results)
                found = True
                yield data_source.dirname(), values
            if not found:
                raise FileNotFoundError(f"No such file or directory: {source}")


def ls_indexed_storages(catalog: "Catalog", long: bool = False) -> Iterator[str]:
    from dvcx.node import long_line_str

    storage_uris = catalog.ls_storage_uris()
    if long:
        for uri in storage_uris:
            # TODO: add Storage.created so it can be used here
            yield long_line_str(uri, None, "")
    else:
        yield from storage_uris


def ls_local(
    sources,
    long: bool = False,
    catalog: Optional["Catalog"] = None,
    client_config=None,
    **kwargs,
):
    if catalog is None:
        from .catalog import get_catalog

        catalog = get_catalog(client_config=client_config)
    if sources:
        actual_sources = list(ls_urls(sources, catalog=catalog, long=long, **kwargs))
        if len(actual_sources) == 1:
            for _, entries in actual_sources:
                for entry in entries:
                    print(format_ls_entry(entry))
        else:
            first = True
            for source, entries in actual_sources:
                # print a newline between directory listings
                if first:
                    first = False
                else:
                    print()
                if source:
                    print(f"{source}:")
                for entry in entries:
                    print(format_ls_entry(entry))
    else:
        for entry in ls_indexed_storages(catalog, long=long):
            print(format_ls_entry(entry))


def format_ls_entry(entry: str) -> str:
    if entry.endswith("/") or not entry:
        entry = shlex.quote(entry[:-1])
        return f"{entry}/"
    return shlex.quote(entry)


def ls_remote(
    url: str,
    username: str,
    token: str,
    paths: Iterable[str],
    long: bool = False,
):
    from dvcx.node import long_line_str
    from dvcx.remote.studio import StudioClient

    client = StudioClient(url, username, token)
    first = True
    for path, response in client.ls(paths):
        if not first:
            print()
        if not response.ok or response.data is None:
            print(f"{path}:\n  Error: {response.message}\n")
            continue

        print(f"{path}:")
        if long:
            for row in response.data:
                entry = long_line_str(
                    row["name"] + ("/" if row["dir_type"] else ""),
                    row["last_modified"],
                    row["owner_name"],
                )
                print(format_ls_entry(entry))
        else:
            for row in response.data:
                entry = row["name"] + ("/" if row["dir_type"] else "")
                print(format_ls_entry(entry))
        first = False


def ls(
    sources,
    long: bool = False,
    remote: str = "",
    config: Optional[Mapping[str, str]] = None,
    **kwargs,
):
    if config is None:
        from .config import get_remote_config, read_config

        config = get_remote_config(read_config(DVCXDir.find().root), remote=remote)
    remote_type = config["type"]
    if remote_type == "local":
        ls_local(sources, long=long, **kwargs)
    else:
        ls_remote(
            config["url"],
            config["username"],
            config["token"],
            sources,
            long=long,
        )


def ls_datasets(catalog: "Catalog"):
    for d in catalog.ls_datasets():
        for v in d.versions:
            print(f"{d.name} (v{v.version})")


def ls_dataset_rows(catalog: "Catalog", name: str, version: int):
    for row in catalog.ls_dataset_rows(name, version):
        entry = row.name + ("/" if row.dir_type else "")  # type: ignore[attr-defined]
        print(format_ls_entry(entry))


def rm_dataset(
    catalog: "Catalog",
    name: str,
    version: Optional[int] = None,
    force: Optional[bool] = False,
):
    catalog.remove_dataset(name, version=version, force=force)


def dataset_stats(
    catalog: "Catalog",
    name: str,
    version: int,
    show_bytes=False,
    si=False,
):
    stats = catalog.dataset_stats(name, version)

    if stats:
        print(f"Number of objects: {stats.num_objects}")
        if show_bytes:
            print(f"Total objects size: {stats.size}")
        else:
            print(f"Total objects size: {utils.sizeof_fmt(stats.size, si=si): >7}")


def du(catalog: "Catalog", sources, show_bytes=False, si=False, **kwargs):
    for path, size in catalog.du(sources, **kwargs):
        if show_bytes:
            print(f"{size} {path}")
        else:
            print(f"{utils.sizeof_fmt(size, si=si): >7} {path}")


def index(
    catalog: "Catalog",
    sources,
    **kwargs,
):
    catalog.index(sources, **kwargs)


def show(
    catalog: "Catalog",
    name: str,
    version: Optional[int] = None,
    limit: int = 10,
    offset: int = 0,
    columns: Sequence[str] = (),
    no_collapse: bool = False,
) -> None:
    from dvcx.query import DatasetQuery
    from dvcx.utils import show_records

    if columns:
        columns = ("id", *columns)
    query = (
        DatasetQuery(name=name, version=version, catalog=catalog)
        .select(*columns)
        .limit(limit)
        .offset(offset)
    )
    records = query.to_records()
    show_records(records, collapse_columns=not no_collapse)


def query(
    catalog: "Catalog",
    script: str,
    dataset_name: Optional[str] = None,
    parallel: Optional[int] = None,
    limit: int = 10,
    offset: int = 0,
    columns: Optional[list[str]] = None,
    no_collapse: bool = False,
    params: Optional[dict[str, str]] = None,
) -> None:
    from dvcx.utils import show_records

    with open(script, encoding="utf-8") as f:
        script_content = f.read()

    if parallel is not None:
        # This also sets this environment variable for any subprocesses
        os.environ["DVCX_SETTINGS_PARALLEL"] = str(parallel)

    result = catalog.query(
        script_content,
        save_as=dataset_name,
        preview_limit=limit,
        preview_offset=offset,
        preview_columns=columns,
        capture_output=False,
        params=params,
    )
    show_records(result.preview, collapse_columns=not no_collapse)


def clear_cache(catalog: "Catalog"):
    catalog.cache.clear()


def garbage_collect(catalog: "Catalog"):
    temp_tables = catalog.get_temp_table_names()
    if not temp_tables:
        print("Nothing to clean up.")
    else:
        print(f"Garbage collecting {len(temp_tables)} tables.")
        catalog.cleanup_temp_tables(temp_tables)


def completion(shell: str) -> str:
    return shtab.complete(
        get_parser(),
        shell=shell,
    )


def main(argv: Optional[list[str]] = None) -> int:  # noqa: C901, PLR0911, PLR0912, PLR0915
    # Required for Windows multiprocessing support
    freeze_support()

    parser = get_parser()
    args = parser.parse_args(argv)

    if args.internal_run_udf:
        from dvcx.query.dispatch import udf_entrypoint

        return udf_entrypoint()

    if args.internal_run_udf_worker:
        from dvcx.query.dispatch import udf_worker_entrypoint

        return udf_worker_entrypoint()

    if args.command is None:
        parser.print_help()
        return 1

    from .catalog import get_catalog

    logger.addHandler(logging.StreamHandler())
    logging_level = get_logging_level(args)
    logger.setLevel(logging_level)

    client_config = {
        "aws_endpoint_url": args.aws_endpoint_url,
        "anon": args.anon,
    }

    if args.debug_sql:
        # This also sets this environment variable for any subprocesses
        os.environ["DEBUG_SHOW_SQL_QUERIES"] = "True"

    try:
        catalog = get_catalog(client_config=client_config)
        if args.command == "cp":
            catalog.cp(
                args.sources,
                args.output,
                force=bool(args.force),
                update=bool(args.update),
                recursive=bool(args.recursive),
                edvcx_file=None,
                edvcx_only=False,
                no_edvcx_file=True,
                no_glob=args.no_glob,
                ttl=args.ttl,
            )
        elif args.command == "clone":
            catalog.clone(
                args.sources,
                args.output,
                force=bool(args.force),
                update=bool(args.update),
                recursive=bool(args.recursive),
                no_glob=args.no_glob,
                ttl=args.ttl,
                no_cp=args.no_cp,
                edvcx=args.edvcx,
                edvcx_file=args.edvcx_file,
            )
        elif args.command == "pull":
            catalog.pull_dataset(
                args.dataset,
                args.output,
                no_cp=args.no_cp,
                force=bool(args.force),
                edvcx=args.edvcx,
                edvcx_file=args.edvcx_file,
            )
        elif args.command == "edit-dataset":
            catalog.edit_dataset(
                args.name,
                description=args.description,
                new_name=args.new_name,
                labels=args.labels,
            )
        elif args.command == "merge-datasets":
            catalog.merge_datasets(
                catalog.get_dataset(args.src),
                catalog.get_dataset(args.dst),
                args.src_version,
                dst_version=args.dst_version,
            )
        elif args.command == "ls":
            ls(
                args.sources,
                long=bool(args.long),
                remote=args.remote,
                ttl=args.ttl,
                update=bool(args.update),
                client_config=client_config,
            )
        elif args.command == "ls-datasets":
            ls_datasets(catalog)
        elif args.command == "ls-dataset-rows":
            ls_dataset_rows(catalog, args.name, args.version)
        elif args.command == "show":
            show(
                catalog,
                args.name,
                args.version,
                limit=args.limit,
                offset=args.offset,
                columns=args.columns,
                no_collapse=args.no_collapse,
            )
        elif args.command == "rm-dataset":
            rm_dataset(catalog, args.name, version=args.version, force=args.force)
        elif args.command == "dataset-stats":
            dataset_stats(
                catalog,
                args.name,
                args.version,
                show_bytes=args.bytes,
                si=args.si,
            )
        elif args.command == "du":
            du(
                catalog,
                args.sources,
                show_bytes=args.bytes,
                depth=args.depth,
                si=args.si,
                ttl=args.ttl,
                update=bool(args.update),
                client_config=client_config,
            )
        elif args.command == "find":
            results_found = False
            for result in catalog.find(
                args.sources,
                ttl=args.ttl,
                update=bool(args.update),
                names=args.name,
                inames=args.iname,
                paths=args.path,
                ipaths=args.ipath,
                size=args.size,
                typ=args.type,
                columns=args.columns,
            ):
                print(result)
                results_found = True
            if not results_found:
                print("No results")
        elif args.command == "index":
            index(
                catalog,
                args.sources,
                ttl=args.ttl,
                update=bool(args.update),
            )
        elif args.command == "completion":
            print(completion(args.shell))
        elif args.command == "find-stale-storages":
            catalog.find_stale_storages()
        elif args.command == "query":
            query(
                catalog,
                args.script,
                dataset_name=args.dataset_name,
                parallel=args.parallel,
                limit=args.limit,
                offset=args.offset,
                columns=args.columns,
                no_collapse=args.no_collapse,
                params=args.param,
            )
        elif args.command == "apply-udf":
            catalog.apply_udf(
                args.udf, args.source, args.target, args.parallel, args.udf_params
            )
        elif args.command == "clear-cache":
            clear_cache(catalog)
        elif args.command == "gc":
            garbage_collect(catalog)
        else:
            print(f"invalid command: {args.command}", file=sys.stderr)
            return 1
        return 0
    except BrokenPipeError:
        # Python flushes standard streams on exit; redirect remaining output
        # to devnull to avoid another BrokenPipeError at shutdown
        # See: https://docs.python.org/3/library/signal.html#note-on-sigpipe
        devnull = os.open(os.devnull, os.O_WRONLY)
        os.dup2(devnull, sys.stdout.fileno())
        return 141  # 128 + 13 (SIGPIPE)
    except (KeyboardInterrupt, Exception) as exc:
        if isinstance(exc, KeyboardInterrupt):
            msg = "Operation cancelled by the user"
        else:
            msg = str(exc)
        print("Error:", msg, file=sys.stderr)
        if logging_level <= logging.DEBUG:
            traceback.print_exception(
                type(exc),
                exc,
                exc.__traceback__,
                file=sys.stderr,
            )
        if args.pdb:
            import pdb  # noqa: T100

            pdb.post_mortem()
        return 1
